import unittest
import shutil
import tempfile
import os
import numpy as np

# VTK imports:
from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa

# Functionality to test:
from .delimited import *
from .binaries import *

RTOL = 0.000001

class TestDelimitedTextReader(unittest.TestCase):
    """
    Test the `DelimitedTextReader`: A widely used base class
    """
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.commafname = os.path.join(self.test_dir, 'comma.txt')
        self.tabfname = os.path.join(self.test_dir, 'tab.txt')
        # Make a temporary delimited text file to test:
        lines = ['This is a header line to skip',
                 'int,str,float ! Comment,this line has the data array names',
                 '5,foo,6.9',
                 '1,bar,8.5 ! another comment',
                 '3,oof,7.7'
                 ]
        # Append newlines
        lines = [ln+'\n' for ln in lines]
        # Now write contents to files
        f = open(self.commafname, 'w')
        f.writelines(lines)
        f.close()
        f = open(self.tabfname, 'w')
        f.writelines([ln.replace(',', '\t') for ln in lines])
        f.close()
        return

    def tearDown(self):
        # Remove the test data directory after the test
        shutil.rmtree(self.test_dir)

    ###########################################

    def _check_shape(self, table):
        # Check number of rows
        self.assertEqual(table.GetNumberOfRows(), 3)
        # Check number of columns:
        self.assertEqual(table.GetNumberOfColumns(), 3)
        return

    def _check_titles(self, table, titles=['int', 'str', 'float']):
        # Check data array names:
        self.assertEqual(table.GetColumnName(0), titles[0])
        self.assertEqual(table.GetColumnName(1), titles[1])
        self.assertEqual(table.GetColumnName(2), titles[2])
        return

    def _check_array_types(self, table, types=[int, float, float]):
        # Check data array types:
        typ = table.GetColumn(0).GetDataType()
        self.assertEqual(typ, nps.get_vtk_array_type(int))
        typ = table.GetColumn(1).GetDataType()
        self.assertEqual(typ, 13)
        typ = table.GetColumn(2).GetDataType()
        self.assertEqual(typ, nps.get_vtk_array_type(float))
        return

    def _check_array_values(self, table):
        # Check data array values:
        wt = dsa.WrapDataObject(table)
        arr = wt.RowData[0]
        self.assertEqual(arr[0], 5)
        self.assertEqual(arr[1], 1)
        self.assertEqual(arr[2], 3)
        arr = table.GetColumn(1)
        self.assertEqual(arr.GetValue(0), 'foo')
        self.assertEqual(arr.GetValue(1), 'bar')
        self.assertEqual(arr.GetValue(2), 'oof')
        arr = wt.RowData[2]
        self.assertEqual(arr[0], 6.9)
        self.assertEqual(arr[1], 8.5)
        self.assertEqual(arr[2], 7.7)
        return


    # TODO: check timesteps!


    def test_comma_read(self):
        """`DelimitedTextReader`: comma delimited file"""
        reader = DelimitedTextReader()
        reader.AddFileName(self.commafname)
        reader.SetDelimiter(',')
        reader.SetSkipRows(1)
        reader.SetComments('!')
        reader.Update()
        table = reader.GetOutput()
        self._check_shape(table)
        self._check_titles(table)
        self._check_array_types(table)
        self._check_array_values(table)
        return

    def test_tab_read(self):
        """`DelimitedTextReader`: tab delimited file"""
        reader = DelimitedTextReader()
        reader.AddFileName(self.tabfname)
        reader.SetUseTab(True)
        reader.SetSkipRows(1)
        reader.SetComments('!')
        reader.Update()
        table = reader.GetOutput()
        self._check_shape(table)
        self._check_titles(table)
        self._check_array_types(table)
        self._check_array_values(table)
        return

    def test_no_titles(self):
        """`DelimitedTextReader`: file without headers"""
        reader = DelimitedTextReader()
        reader.AddFileName(self.commafname)
        reader.SetDelimiter(',')
        reader.SetSkipRows(2)
        reader.SetHasTitles(False)
        reader.SetComments('!')
        reader.Update()
        table = reader.GetOutput()
        self._check_shape(table)
        self._check_titles(table, titles=['Field 0', 'Field 1', 'Field 2'])
        self._check_array_types(table)
        self._check_array_values(table)
        return





###############################################################################

class TestXYZTextReader(unittest.TestCase):
    """
    Test the `XYZTextReader`
    """
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.fname = os.path.join(self.test_dir, 'test.xyz')
        # Make a temporary file to test:
        self.nrows = 100
        self.ncols = 8 # LEAVE ALONE
        self.header = 'X, dx, Y, dy, Z, dz, approximate distance, cell index'
        self.data = np.random.random((self.nrows, self.ncols))
        np.savetxt(self.fname, self.data,
                   header=self.header,
                   comments='! ',
                   fmt='%.6e'
                   )
        reader = XYZTextReader()
        reader.AddFileName(self.fname)
        reader.Update()
        self.TABLE = reader.GetOutput()
        return

    def tearDown(self):
        # Remove the test data directory after the test
        shutil.rmtree(self.test_dir)

    ###########################################

    def test_data_aray_titles(self):
        """`XYZTextReader`: check data array names"""
        titles = self.header.split(', ')
        for i in range(self.ncols):
            self.assertEqual(self.TABLE.GetColumnName(i), titles[i])
        return

    def test_data_fidelity(self):
        """`XYZTextReader`: check data fidelity"""
        titles = self.header.split(', ')
        for i in range(self.ncols):
            arr = nps.vtk_to_numpy(self.TABLE.GetColumnByName(titles[i]))
            self.assertTrue(np.allclose(self.data[:,i], arr, rtol=RTOL))
        return

    def test_shape(self):
        """`XYZTextReader`: data table shape"""
        self.assertEqual(self.TABLE.GetNumberOfRows(), self.nrows)
        self.assertEqual(self.TABLE.GetNumberOfColumns(), self.ncols)
        return

###############################################################################


class TestPackedBinariesReader(unittest.TestCase):
    """
    Test the `PackedBinariesReader`
    """
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.n = 100
        return

    def tearDown(self):
        # Remove the test data directory after the test
        shutil.rmtree(self.test_dir)

    ###########################################

    def _check_data(self, table, data):
        arr = nps.vtk_to_numpy(table.GetColumn(0))
        self.assertTrue(np.allclose(data, arr, rtol=RTOL))
        return arr

    ###########################################

    def test_floats(self):
        """`PackedBinariesReader`: floats"""
        # Make data and write out
        dtype = np.dtype('f')
        arr = np.array(np.random.random(self.n), dtype=dtype)
        fname = os.path.join(self.test_dir, 'test.bin')
        arr.tofile(fname)
        # Set up reader
        reader = PackedBinariesReader()
        reader.AddFileName(fname)
        reader.SetDataType('f')
        reader.SetDataName('Test Data')
        # Perfrom Read
        reader.Update()
        table = reader.GetOutput()
        # Check output
        self.assertEqual(table.GetColumnName(0), 'Test Data')
        self._check_data(table, arr)
        return

    def test_doubles(self):
        """`PackedBinariesReader`: doubles"""
        # Make data and write out
        dtype = np.dtype('d')
        arr = np.array(np.random.random(self.n), dtype=dtype)
        fname = os.path.join(self.test_dir, 'test.bin')
        arr.tofile(fname)
        # Set up reader
        reader = PackedBinariesReader()
        reader.AddFileName(fname)
        reader.SetDataType('d')
        # Perfrom Read
        reader.Update()
        table = reader.GetOutput()
        # Check output
        self._check_data(table, arr)
        return

    def test_ints(self):
        """`PackedBinariesReader`: ints"""
        # Make data and write out
        dtype = np.dtype('i')
        arr = np.array(np.random.random(self.n), dtype=dtype)
        fname = os.path.join(self.test_dir, 'test.bin')
        arr.tofile(fname)
        # Set up reader
        reader = PackedBinariesReader()
        reader.AddFileName(fname)
        reader.SetDataType(2) # 'i' test that sending an int choice works
        # Perfrom Read
        reader.Update()
        table = reader.GetOutput()
        # Check output
        self._check_data(table, arr)
        return

    def test_endian_big(self):
        """`PackedBinariesReader`: floats with big-endianness"""
        # Make data and write out
        dtype = np.dtype('>f')
        arr = np.asarray(np.random.random(self.n), dtype=dtype)
        fname = os.path.join(self.test_dir, 'test.bin')
        arr.tofile(fname)
        # Set up reader
        reader = PackedBinariesReader()
        reader.AddFileName(fname)
        reader.SetDataType('f')
        reader.SetEndian('>')
        # Perfrom Read
        reader.Update()
        table = reader.GetOutput()
        # Check output
        self._check_data(table, arr)
        return

    def test_endian_little(self):
        """`PackedBinariesReader`: floats with little-endianness"""
        # Make data and write out
        dtype = np.dtype('<f')
        arr = np.array(np.random.random(self.n), dtype=dtype)
        fname = os.path.join(self.test_dir, 'test.bin')
        arr.tofile(fname)
        # Set up reader
        reader = PackedBinariesReader()
        reader.AddFileName(fname)
        reader.SetDataType('f')
        reader.SetEndian(1) # '<' test that sending an int choice works
        # Perfrom Read
        reader.Update()
        table = reader.GetOutput()
        # Check output
        self._check_data(table, arr)
        return

###############################################################################

class TestMadagascarReader(unittest.TestCase):
    """
    Test the `MadagascarReader`
    Does not test inherrited functionality
    """

    def tearDown(self):
        # Remove the test data directory after the test
        shutil.rmtree(self.test_dir)

    ###########################################

    def test_data_fidelity(self):
        """`MadagascarReader`: Check data fidelity"""
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.n = 100
        # Make data and write out
        dtype = np.dtype('f')
        self.data = np.array(np.random.random(self.n), dtype=dtype)
        fname = os.path.join(self.test_dir, 'test.rsf')
        # Write ascii header
        lines = ['hello\n']*10
        with open(fname, 'w') as f:
            f.writelines(lines)
        # Write data
        with open(fname, 'ab') as f:
            f.write(b'\014\014\004') # The control sequence
            self.data.tofile(f)
        # Set up reader
        reader = MadagascarReader()
        reader.AddFileName(fname)
        reader.SetDataType('f')
        reader.SetDataName('Test Data')
        self.assertEqual(reader.GetDataName(), 'Test Data')
        # Perfrom Read
        reader.Update()
        table = reader.GetOutput()
        arr = nps.vtk_to_numpy(table.GetColumn(0))
        self.assertTrue(np.allclose(self.data, arr))#, rtol=0.0001))
        self.assertEqual(table.GetColumnName(0), 'Test Data')
        return


###############################################################################


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np

# VTK imports:
import vtk
from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa

# Functionality to test:
from .earth import *
from .evenModel import *
from .oddModel import *

###############################################################################

class TestEarthSource(unittest.TestCase):
    """
    Test the `EarthSource` source
    """

    def test_generation(self):
        """`EarthSource`: make sure works."""
        s = EarthSource()
        s.SetRadius(6000.0)
        s.Update()
        e = s.GetOutput()
        self.assertIsNotNone(e)
        self.assertTrue(isinstance(e, vtk.vtkPolyData))
        return


###############################################################################

class TestCreateUniformGrid(unittest.TestCase):
    """
    Test the `CreateUniformGrid` source
    """

    def test_generation(self):
        """`CreateUniformGrid`: make sure works."""
        g = CreateUniformGrid()
        g.SetExtent(20, 10, 35)
        g.SetOrigin(33.3, 45.5, 6.6)
        g.SetSpacing(5,1,7)
        g.Update()
        grid = g.GetOutput()
        self.assertIsNotNone(grid)
        self.assertEqual(grid.GetNumberOfCells(), (19*9*34))
        self.assertEqual(grid.GetNumberOfPoints(), (20*10*35))
        bounds = grid.GetBounds()
        tb = (33.3, 33.3+(19*5), 45.5, 45.5+(9*1), 6.6, 6.6+(34*7))
        self.assertEqual(bounds, tb)
        return


###############################################################################

class TestCreateEvenRectilinearGrid(unittest.TestCase):
    """
    Test the `CreateEvenRectilinearGrid` source
    """

    def test_generation(self):
        """`CreateEvenRectilinearGrid`: make sure works."""
        g = CreateEvenRectilinearGrid()
        g.SetExtent(20, 10, 35)
        g.SetXRange(0, 100)
        g.SetYRange(-100, 0)
        g.SetZRange(10, 50)
        g.Update()
        grid = g.GetOutput()
        self.assertIsNotNone(grid)
        self.assertEqual(grid.GetNumberOfCells(), (20*10*35))
        self.assertEqual(grid.GetNumberOfPoints(), (21*11*36))
        bounds = grid.GetBounds()
        tb = (0, 100, -100, 0, 10, 50)
        self.assertEqual(bounds, tb)


###############################################################################

class TestCreateTensorMesh(unittest.TestCase):
    """
    Test the `CreateTensorMesh` source
    """

    def test_generation(self):
        """`CreateTensorMesh`: make sure works."""
        g = CreateTensorMesh()
        g.SetXCellsStr('200 100 50 30*50.0 50 100 200')
        g.SetYCellsStr('200 100 50 25*50.0 50 100 200')
        g.SetZCellsStr('25*25.0 50 100 200')
        g.SetOrigin(-300.0, -450.0, 10.0)
        g.Update()
        grid = g.GetOutput()
        self.assertIsNotNone(grid)
        self.assertEqual(grid.GetNumberOfCells(), (36*31*28))
        bounds = grid.GetBounds()
        tb = (-300.0, 1900, -450, 1500, -965, 10)
        self.assertEqual(bounds, tb)


###############################################################################



all = [
    'ubcMeshReaderBase',
    'ModelAppenderBase',
]

from .. import _helpers
from .. import base
# Outside Imports:
import numpy as np
import vtk

###############################################################################


# UBC Mesh Reader Base
class ubcMeshReaderBase(base.TwoFileReaderBase):
    """A base class for the UBC mesh readers
    """
    __displayname__ = 'UBC Mesh Reader Base'
    __type__ = 'base'
    def __init__(self, nOutputPorts=1, outputType='vtkUnstructuredGrid', **kwargs):
        base.TwoFileReaderBase.__init__(self,
            nOutputPorts=nOutputPorts, outputType=outputType,
            **kwargs)
        self.__dataname = 'Data'
        # For keeping track of type (2D vs 3D)
        self.__sizeM = None


    def Is3D(self):
        return self.__sizeM.shape[0] >= 3

    def Is2D(self):
        return self.__sizeM.shape[0] == 1

    @staticmethod
    def _ubcMesh2D_part(FileName):
        # This is a helper method to read file contents of mesh
        try:
            fileLines = np.genfromtxt(FileName, dtype=str, delimiter='\n', comments='!')
        except (FileNotFoundError, OSError) as fe:
            raise _helpers.PVGeoError(str(fe))

        def _genTup(sft, n):
            # This reads in the data for a dimension
            pts = []
            disc = []
            for i in range(n):
                ln = fileLines[i+sft].split('!')[0].split()
                if i is 0:
                    o = ln[0]
                    pts.append(o)
                    ln = [ln[1],ln[2]]
                pts.append(ln[0])
                disc.append(ln[1])
            return pts, disc

        # Get the number of lines for each dimension
        nx = int(fileLines[0].split('!')[0])
        nz = int(fileLines[nx+1].split('!')[0])

        # Get the origins and tups for both dimensions
        xpts, xdisc = _genTup(1, nx)
        zpts, zdisc = _genTup(2+nx, nz)

        return xpts, xdisc, zpts, zdisc

    def _ReadExtent(self):
        """Reads the mesh file for the UBC 2D/3D Mesh or OcTree format to get output extents. Computationally inexpensive method to discover whole output extent.

        Return:
            tuple : This returns a tuple of the whole extent for the grid to be made of the input mesh file (0,n1-1, 0,n2-1, 0,n3-1). This output should be directly passed to set the whole output extent.

        """
        # Read the mesh file as line strings, remove lines with comment = !
        v = np.array(np.__version__.split('.')[0:2], dtype=int)
        FileName = self.GetMeshFileName()
        try:
            if v[0] >= 1 and v[1] >= 10:
                # max_rows in numpy versions >= 1.10
                msh = np.genfromtxt(FileName, delimiter='\n', dtype=np.str,comments='!', max_rows=1)
            else:
                # This reads whole file :(
                msh = np.genfromtxt(FileName, delimiter='\n', dtype=np.str, comments='!')[0]
        except (FileNotFoundError, OSError) as fe:
            raise _helpers.PVGeoError(str(fe))
        # Fist line is the size of the model
        self.__sizeM = np.array(msh.ravel()[0].split(), dtype=int)
        # Check if the mesh is a UBC 2D mesh
        if self.__sizeM.shape[0] == 1:
            # Read in data from file
            xpts, xdisc, zpts, zdisc = ubcMeshReaderBase._ubcMesh2D_part(FileName)
            nx = np.sum(np.array(xdisc,dtype=int))+1
            nz = np.sum(np.array(zdisc,dtype=int))+1
            return (0,nx, 0,1, 0,nz)
        # Check if the mesh is a UBC 3D mesh or OcTree
        elif self.__sizeM.shape[0] >= 3:
            # Get mesh dimensions
            dim = self.__sizeM[0:3]
            ne,nn,nz = dim[0], dim[1], dim[2]
            return (0,ne, 0,nn, 0,nz)
        else:
            raise _helpers.PVGeoError('File format not recognized')


    @staticmethod
    def ubcModel3D(FileName):
        """Reads the 3D model file and returns a 1D NumPy float array. Use the PlaceModelOnMesh() method to associate with a grid.

        Args:
            FileName (str) : The model file name(s) as an absolute path for the input model file in UBC 3D Model Model Format. Also accepts a `list` of string file names.

        Return:
            np.array : Returns a NumPy float array that holds the model data read from the file. Use the ``PlaceModelOnMesh()`` method to associate with a grid. If a list of file names is given then it will return a dictionary of NumPy float array with keys as the basenames of the files.
        """
        if type(FileName) is list:
            out = {}
            for f in FileName:
                out[os.path.basename(f)] = TensorMeshReader.ubcModel3D(f)
            return out
        try:
            fileLines = np.genfromtxt(FileName, dtype=str, delimiter='\n', comments='!')
        except (FileNotFoundError, OSError) as fe:
            raise _helpers.PVGeoError(str(fe))
        data = np.genfromtxt((line.encode('utf8') for line in fileLines), dtype=np.float)
        return data


    def SetDataName(self, name):
        """Set te data array name for the model data on the output grid.
        """
        if self.__dataname != name:
            self.__dataname = name
            self.Modified(readAgainMesh=False, readAgainModels=False)

    def GetDataName(self):
        return self.__dataname




###############################################################################


# UBC Model Appender Base
class ModelAppenderBase(base.AlgorithmBase):
    """A base class for create mesh-model appenders on the UBC Mesh formats
    """
    __displayname__ = 'Model Appender Base'
    __type__ = 'base'
    def __init__(self, inputType='vtkRectilinearGrid', outputType='vtkRectilinearGrid', **kwargs):
        base.AlgorithmBase.__init__(self,
            nInputPorts=1, inputType=inputType,
            nOutputPorts=1, outputType=outputType)
        self._modelFileNames = kwargs.get('modelfiles', [])
        self._dataname = kwargs.get('dataname', 'Appended Data')
        self._models = []
        self.__needToRead = True
        self._is3D = None
        # For the VTK/ParaView pipeline
        self.__dt = kwargs.get('dt', 1.0)
        self.__timesteps = None
        self.__inTimesteps = None

    def __SetInputTimesteps(self):
        ints = _helpers.GetInputTimeSteps(self)
        self.__inTimesteps = ints if ints is not None else []
        return self.__inTimesteps

    def NeedToRead(self, flag=None):
        """Ask self if the reader needs to read the files again

        Args:
            flag (bool): if the flag is set then this method will set the read status

        Return:
            bool: The status of the reader aspect of the filter.
        """
        if flag is not None and isinstance(flag, (bool, int)):
            self.__needToRead = flag
            self.__UpdateTimeSteps()
        return self.__needToRead

    def Modified(self, readAgain=True):
        """Call modified if the files needs to be read again again.
        """
        if readAgain: self.__needToRead = readAgain
        base.AlgorithmBase.Modified(self)

    def __UpdateTimeSteps(self):
        """For internal use only: appropriately sets the timesteps.
        """
        if len(self._modelFileNames) > 0 and len(self._modelFileNames) > len(self.__inTimesteps):
            self.__timesteps = _helpers.UpdateTimeSteps(self, self._modelFileNames, self.__dt)
        # Just use input's time steps which is set by pipeline
        return 1

    def _ReadUpFront(self):
        raise NotImpelementedError()

    def _PlaceOnMesh(self, output, idx=0):
        raise NotImplementedError()


    def RequestData(self, request, inInfo, outInfo):
        """DO NOT OVERRIDE
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        output = self.GetOutputData(outInfo, 0)
        output.DeepCopy(pdi) # ShallowCopy if you want changes to propagate upstream
        # Get requested time index
        i = _helpers.GetRequestedTime(self, outInfo)
        # Perfrom task:
        if self.__needToRead:
            self._ReadUpFront()
        # Place the model data for given timestep onto the mesh
        if len(self._models) > i:
            self._PlaceOnMesh(output, idx=i)
        return 1

    def RequestInformation(self, request, inInfo, outInfo):
        """DO NOT OVERRIDE
        """
        self.__SetInputTimesteps()
        self.__UpdateTimeSteps()
        pdi = self.GetInputData(inInfo, 0, 0)
        # Determine if 2D or 3D and read
        if isinstance(pdi, vtk.vtkRectilinearGrid) and pdi.GetExtent()[3] == 1:
            self._is3D = False
        else:
            self._is3D = True
        return 1

    #### Setters and Getters ####

    def HasModels(self):
        return len(self._modelFileNames) > 0

    def GetTimestepValues(self):
        """Use this in ParaView decorator to register timesteps.
        """
        if self.__timesteps is None: self.__timesteps = self.__SetInputTimesteps()
        return self.__timesteps.tolist() if self.__timesteps is not None else None

    def ClearModels(self):
        """Use to clear data file names.
        """
        self._modelFileNames = []
        self._models = []
        self.Modified(readAgain=True)

    def AddModelFileName(self, fname):
        """Use to set the file names for the reader. Handles singlt string or list of strings.
        """
        if fname is None:
            return # do nothing if None is passed by a constructor on accident
        if isinstance(fname, list):
            for f in fname:
                self.AddModelFileName(f)
            self.Modified()
        elif fname not in self._modelFileNames:
            self._modelFileNames.append(fname)
            self.Modified()
        return 1

    def GetModelFileNames(self, idx=None):
        """Returns the list of file names or given and index returns a specified timestep's filename.
        """
        if idx is None or not self.HasModels():
            return self._modelFileNames
        return self._modelFileNames[idx]

    def SetDataName(self, name):
        if self._dataname != name:
            self._dataname = name
            self.Modified(readAgain=False)



import unittest
import shutil
import tempfile
import os
import numpy as np

# VTK imports:
from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa

# Functionality to test:
from .tensor_mesh import *
from .octree import *

RTOL = 0.000001

###############################################################################

class ubcMeshTesterBase(unittest.TestCase):

    def _check_shape(self, grid):
        self.assertEqual(grid.GetExtent(), self.extent)
        self.assertEqual(grid.GetNumberOfCells(), self.extent[1]*self.extent[3]*self.extent[5])
        return

    def _check_data(self, grid, data):
        arr = nps.vtk_to_numpy(grid.GetCellData().GetArray(0))
        self.assertTrue(np.allclose(data, arr, rtol=RTOL))
        return

    def _check_spatial_reference(self, grid):
        bounds = grid.GetBounds()
        corner = (bounds[0], bounds[2], bounds[5])
        self.assertEqual(corner, self.origin)
        return

###############################################################################

class Test3DTensorMeshReader(ubcMeshTesterBase):
    """
    Test the `TensorMeshReader` and `TensorMeshAppender` for 3D data
    """
    def _write_mesh(self):
        fname = os.path.join(self.test_dir, 'test.msh')
        with open(fname, 'w') as f:
            f.write('%d %d %d\n' % self.shape)
            f.write('%d %d %d\n' % self.origin)
            f.write('%s\n' % self.xCells)
            f.write('%s\n' % self.yCells)
            f.write('%s\n' % self.zCells)
        return fname

    def _write_model(self, fname='test.mod'):
        fname = os.path.join(self.test_dir, fname)
        model = np.random.random(self.n)
        np.savetxt(fname, model, delimiter=' ', comments='! ')
        model = np.reshape(model, self.shape)
        model = np.swapaxes(model,0,1)
        model = np.swapaxes(model,0,2)
        # Now reverse Z axis
        model = model[::-1,:,:] # Note it is in Fortran ordering
        model = model.flatten()
        return fname, model

    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.origin = (-350, -400, 0)
        self.xCells = '200 100 50 20*50.0 50 100 200'
        self.yCells = '200 100 50 21*50.0 50 100 200'
        self.zCells = '20*25.0 50 100 200'
        self.shape = (26, 27, 23)
        self.n = self.shape[0]*self.shape[1]*self.shape[2]
        self.extent = (0, self.shape[0], 0, self.shape[1], 0, self.shape[2])
        self.dataName = 'foo'
        ##### Now generate output for testing ####
        # Produce data and write out files:
        meshname = self._write_mesh()
        modname, self.data = self._write_model()
        # Set up the reader:
        reader = TensorMeshReader()
        reader.SetMeshFileName(meshname)
        # Get and test output:
        reader.Update() # Read only mesh upfront
        reader.AddModelFileName(modname)
        reader.SetDataName(self.dataName)
        reader.Update() # Read models upfront
        self.GRID = reader.GetOutput()

    def tearDown(self):
        # Remove the test data directory after the test
        shutil.rmtree(self.test_dir)

    ###########################################

    def test_grid_spatial_reference(self):
        """`TensorMeshReader` 3D: Spatial reference"""
        self._check_spatial_reference(self.GRID)

    def test_grid_shape(self):
        """`TensorMeshReader` 3D: Shape of output grid"""
        self._check_shape(self.GRID)

    def test_grid_data(self):
        """`TensorMeshReader` 3D: Data fidelity"""
        self._check_data(self.GRID, self.data)

    def test_grid_data_name(self):
        """`TensorMeshReader` 3D: Data array name"""
        self.assertEqual(self.GRID.GetCellData().GetArrayName(0), self.dataName)

    def test_model_appender(self):
        """`TensorMeshAppender` 3D: Data array name"""
        modname, appdata = self._write_model('testApp.mod')
        f = TensorMeshAppender()
        f.SetInputDataObject(self.GRID)
        f.AddModelFileName(modname)
        f.SetDataName('appended')
        f.Update()
        output = f.GetOutput()
        self.assertEqual(output.GetCellData().GetNumberOfArrays(), 2)
        self.assertEqual(output.GetCellData().GetArrayName(1), 'appended')

###############################################################################

class Test2DTensorMeshReader(ubcMeshTesterBase):
    """
    Test the `TensorMeshReader` and `TensorMeshAppender` for 2D data
    """

    def _write_mesh(self):
        fname = os.path.join(self.test_dir, 'test.msh')
        with open(fname, 'w') as f:
            f.write(self.mesh)
        return fname

    def _write_model(self, fname='test.mod'):
        fname = os.path.join(self.test_dir, fname)
        model = np.random.random((self.nz,self.nx))
        with open(fname, 'w') as f:
            f.write('%d %d\n' % (self.nx, self.nz))
            for k in range(self.nz):
                for i in range(self.nx):
                    f.write('%.6e ' % model[k,i])
                f.write('\n')
            f.close()
        model = np.reshape(model.flatten(order='F'), self.shape)
        model = np.swapaxes(model,0,1)
        model = np.swapaxes(model,0,2)
        # Now reverse Z axis
        model = model[::-1,:,:] # Note it is in Fortran ordering
        model = model.flatten()
        return fname, model

    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.mesh = """9
 -300.0   -180.0     1
          -130.0     1
          -110.0     1
          -100.0     1
           100.0    40
           110.0     1
           130.0     1
           180.0     1
           300.0     1
14
 -10.0      10.0     5
            22.0     4
            42.0     5
            57.0     3
            63.0     1
            71.0     1
            81.0     1
            95.0     1
           115.0     1
           140.0     1
           170.0     1
           205.0     1
           245.0     1
           300.0     1
"""
        self.origin = (-300, 0, 10)
        self.nx = 48
        self.nz = 27
        self.shape = (self.nx, 1, self.nz)
        self.extent = (0, self.shape[0], 0, self.shape[1], 0, self.shape[2])
        self.dataName = 'foo'
        ##### Now generate output for testing ####
        # Produce data and write out files:
        meshname = self._write_mesh()
        modname, self.data = self._write_model()
        # Set up the reader:
        reader = TensorMeshReader()
        reader.SetMeshFileName(meshname)
        # Get and test output:
        reader.Update() # Test the read up front for the mesh
        reader.AddModelFileName(modname)
        reader.SetDataName(self.dataName)
        reader.Update() # Now read the models upfront
        self.GRID = reader.GetOutput()
        return

    def tearDown(self):
        # Remove the test data directory after the test
        shutil.rmtree(self.test_dir)

    ###########################################

    def test_grid_spatial_reference(self):
        """`TensorMeshReader` 2D: Spatial reference"""
        self._check_spatial_reference(self.GRID)

    def test_grid_shape(self):
        """`TensorMeshReader` 2D: Shape of output grid"""
        self._check_shape(self.GRID)

    def test_grid_data(self):
        """`TensorMeshReader` 2D: Data fidelity"""
        self._check_data(self.GRID, self.data)

    def test_grid_data_name(self):
        """`TensorMeshReader` 2D: Data array name"""
        self.assertEqual(self.GRID.GetCellData().GetArrayName(0), self.dataName)

    def test_model_appender(self):
        """`TensorMeshAppender` 2D: Data array name"""
        modname, appdata = self._write_model('testApp.mod')
        f = TensorMeshAppender()
        f.SetInputDataObject(self.GRID)
        f.AddModelFileName(modname)
        f.SetDataName('appended')
        f.Update()
        output = f.GetOutput()
        self.assertEqual(output.GetCellData().GetNumberOfArrays(), 2)
        self.assertEqual(output.GetCellData().GetArrayName(1), 'appended')


###############################################################################

class TestOcTreeMeshReader(ubcMeshTesterBase):
    """
    Test the `OcTreeReader`
    """

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        treeMesh = """16 16 16
0.0000 0.0000 48.0000
1.000 2.000 3.000
29
1 1 1 8
9 1 1 8
1 9 1 8
9 9 1 8
1 1 9 4
5 1 9 4
9 1 9 8
1 5 9 4
5 5 9 4
1 9 9 8
9 9 9 8
1 1 13 2
3 1 13 2
5 1 13 4
1 3 13 2
3 3 13 2
1 5 13 4
5 5 13 4
1 1 15 1
2 1 15 1
3 1 15 2
1 2 15 1
2 2 15 1
1 3 15 2
3 3 15 2
1 1 16 1
2 1 16 1
1 2 16 1
2 2 16 1
"""
        # Write out mesh file
        fname = os.path.join(self.test_dir, 'octree.msh')
        self.meshFileName = fname
        with open(fname, 'w') as f:
            f.write(treeMesh)


        # write out model file(s)
        self.nt = 5
        self.modelFileNames = ['model%d.mod' % i for i in range(self.nt)]
        self.modelFileNames = [os.path.join(self.test_dir, self.modelFileNames[i]) for i in range(self.nt)]
        self.arrs = [None] * self.nt
        for i in range(self.nt):
            self.arrs[i] = np.random.random(29)
            np.savetxt(self.modelFileNames[i], self.arrs[i], delimiter=' ', comments='! ')
        return

    def tearDown(self):
        # Remove the test data directory after the test
        shutil.rmtree(self.test_dir)

    def reshapeArrs(self, mesh):
        for i in range(self.nt):
            ind_reorder = nps.vtk_to_numpy(
                mesh.GetCellData().GetArray('index_cell_corner'))
            self.arrs[i] = self.arrs[i][ind_reorder]

    def test_simple_octree(self):
        """`OcTreeReader`: simple octree mesh file"""
        reader = OcTreeReader()
        reader.SetMeshFileName(self.meshFileName)

        reader.Update()

        tree = reader.GetOutput()
        self.assertIsNotNone(tree)
        self.assertEqual(tree.GetNumberOfCells(), 29)
        self.assertEqual(tree.GetNumberOfPoints(), 84)

    def test_simple_octree(self):
        """`OcTreeReader`: simple octree mesh with models"""
        reader = OcTreeReader()
        reader.SetMeshFileName(self.meshFileName)
        reader.AddModelFileName(self.modelFileNames)
        reader.SetDataName('foo')

        reader.Update() # Check that normal update works

        tree = reader.GetOutput()
        self.assertIsNotNone(tree)
        self.assertEqual(tree.GetNumberOfCells(), 29)
        self.assertEqual(tree.GetNumberOfPoints(), 84)

        self.reshapeArrs(tree)

        wtree = dsa.WrapDataObject(tree)
        # Now check time series
        for i in range(self.nt):
            reader.UpdateTimeStep(i)
            arr = wtree.CellData['foo']
            self.assertTrue(np.allclose(arr, self.arrs[i], rtol=RTOL))

        return

    def test_model_appender(self):
        """`OcTreeAppender` 2D: Data array name"""
        # Creat a tree mesh to append
        reader = OcTreeReader()
        reader.SetMeshFileName(self.meshFileName)
        reader.AddModelFileName(self.modelFileNames[0])
        reader.SetDataName('Initial Data')
        reader.Update()
        tree = reader.GetOutput()
        self.assertIsNotNone(tree)
        self.assertEqual(tree.GetNumberOfCells(), 29)
        self.assertEqual(tree.GetNumberOfPoints(), 84)

        # Now use the model appender
        f = OcTreeAppender()
        f.SetInputDataObject(tree)
        f.AddModelFileName(self.modelFileNames[1::])
        f.SetDataName('Appended Data')
        f.Update()

        output = f.GetOutput()
        self.assertEqual(output.GetCellData().GetNumberOfArrays(), 3) # remember that the index corner array is added by the reader
        self.assertEqual(output.GetCellData().GetArrayName(2), 'Appended Data')
        self.assertEqual(len(f.GetTimestepValues()), self.nt-1)
        return



__all__ = [
    'ExtractTopography',
]

import vtk
from vtk.util import numpy_support as nps
import numpy as np
from vtk.numpy_interface import dataset_adapter as dsa
from datetime import datetime
# Import Helpers:
from ..base import FilterBase
from .. import _helpers
# NOTE: internal import - from scipy.spatial import cKDTree



###############################################################################

class ExtractTopography(FilterBase):
    """This filter takes two inputs: a gridded data set and a set of points for a Topography source. This will add a boolean data array to the cell data of the input grid on whether that cell should be active (under topographic layer).

    """
    __displayname__ = 'Extract Topography'
    __type__ = 'filter'
    def __init__(self):
        FilterBase.__init__(self,
            nInputPorts=2, inputType='vtkDataObject',
            nOutputPorts=1)

    # CRITICAL for multiple input ports
    def FillInputPortInformation(self, port, info):
        """This simply makes sure the user selects the correct inputs
        """
        typ = 'vtkDataSet'
        if port == 1:
            typ = 'vtkPolyData' # Make sure topography are poly data
        info.Set(self.INPUT_REQUIRED_DATA_TYPE(), typ)
        return 1

    # THIS IS CRUCIAL to preserve data type through filter
    def RequestDataObject(self, request, inInfo, outInfo):
        """Constructs the output data object based on the input data object
        """
        self.OutputType = self.GetInputData(inInfo, 0, 0).GetClassName()
        self.FillOutputPortInformation(0, outInfo.GetInformationObject(0))
        return 1


    @staticmethod
    def __GetVoxelCenter(voxel):
        """Returns tuple for center of Voxel

        Note: The Z-coordinate is at the top of the cell
        """
        bounds = voxel.GetBounds()
        x = (bounds[0]+bounds[1])/2.0
        y = (bounds[2]+bounds[3])/2.0
        z = bounds[5]#(bounds[4]+bounds[5])/2.0
        return (x, y, z)

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        from scipy.spatial import cKDTree # NOTE: Must have SciPy in ParaView
        # Get input/output of Proxy
        igrid = self.GetInputData(inInfo, 0, 0) # Port 0: grid
        topo = self.GetInputData(inInfo, 1, 0) # Port 1: topography
        grid = self.GetOutputData(outInfo, 0)
        # Perfrom task
        grid.DeepCopy(igrid)
        ncells = grid.GetNumberOfCells()
        active = np.zeros((ncells), dtype=int)
        # Now iterate through the cells in the grid and test if they are beneath the topography
        wtopo = dsa.WrapDataObject(topo) # NumPy wrapped points
        topoPts = np.array(wtopo.Points) # New NumPy array of points so we dont destroy input
        tree = cKDTree(topoPts[:, 0:2]) # NOTE: only on the XY plane

        # OPTIMIZE: accelerate this to harness structured coords
        for i in range(ncells):
            voxel = grid.GetCell(i)
            x,y,z = self.__GetVoxelCenter(voxel)
            # Search for same XY point in topoPts
            ptsi = tree.query([[x, y]], k=1)[1]
            if z > topoPts[ptsi[0]][2]:
                active[i] = 0
            else:
                active[i] = 1


        # Now add cell data to output
        data = nps.numpy_to_vtk(num_array=active, deep=True)
        data.SetName('Active Topography')
        grid.GetCellData().AddArray(data)
        return 1

    def Apply(self, data, points):
        self.SetInputDataObject(0, data)
        self.SetInputDataObject(1, points)
        self.Update()
        return self.GetOutput()



"""
This package is a set of module files for each sub-package in PVGeo. Each file here
contains the neccessary wrappings for PVGeo to be used directly in ParaView.
"""
from .PVGeo_Filters import *
from .PVGeo_Grid_Tools import *
from .PVGeo_GSLib import *
from .PVGeo_Model_Builder import *
from .PVGeo_Readers import *
from .PVGeo_Tunneling import *
from .PVGeo_UBC_Tools import *


__author__ = 'Bane Sullivan'
__license__ = 'BSD-3-Clause'
__copyright__ = '2018, Bane Sullivan'
__version__ = '1.1.10'



from __future__ import print_function
from .objs import *
from .axes import *

__displayname__ = 'Visualization'

def hideAll():
    """This hides all sources/filters on the pipeline from the current view
    """
    import paraview.simple as pvs
    for f in pvs.GetSources().values():
        pvs.Hide(f)
    return None

hideAll.__displayname__ = 'Hide All'
hideAll.__type__ = 'macro'



__all__ = [
    'CreateEvenRectilinearGrid',
    'CreateUniformGrid',
]

import vtk
from vtk.util import numpy_support as nps
import numpy as np
#from vtk.numpy_interface import dataset_adapter as dsa
# Import Helpers:
from ..base import AlgorithmBase
from .. import _helpers


def _makeSpatialCellData(nx, ny, nz):
    """Used for testing
    """
    arr = np.fromfunction(lambda k, j, i: k*j*i, (nz, ny, nz))
    return arr.flatten()


class CreateUniformGrid(AlgorithmBase):
    """Create uniform grid (``vtkImageData``)
    """
    __displayname__ = 'Create Uniform Grid'
    __type__ = 'source'
    def __init__(self,
                 extent=[10, 10, 10],
                 spacing=[1.0, 1.0, 1.0],
                 origin=[0.0, 0.0, 0.0]):
        AlgorithmBase.__init__(self,
            nInputPorts=0,
            nOutputPorts=1, outputType='vtkImageData')
        self.__extent = extent
        self.__spacing = spacing
        self.__origin = origin


    def RequestData(self, request, inInfo, outInfo):
        pdo = self.GetOutputData(outInfo, 0)
        nx,ny,nz = self.__extent[0],self.__extent[1],self.__extent[2]
        sx,sy,sz = self.__spacing[0],self.__spacing[1],self.__spacing[2]
        ox,oy,oz = self.__origin[0],self.__origin[1],self.__origin[2]
        # Setup the ImageData
        pdo.SetDimensions(nx, ny, nz)
        pdo.SetOrigin(ox, oy, oz)
        pdo.SetSpacing(sx, sy, sz)
        #pdo.SetExtent(0,nx-1, 0,ny-1, 0,nz-1)
        # Add CELL data
        data = _makeSpatialCellData(nx-1, ny-1, nz-1) # minus 1 b/c cell data not point data
        data = nps.numpy_to_vtk(num_array=data, deep=True)
        data.SetName('Spatial Cell Data')
        # THIS IS CELL DATA! Add the model data to CELL data:
        pdo.GetCellData().AddArray(data)
        # Add Point data
        data = _makeSpatialCellData(nx, ny, nz)
        data = nps.numpy_to_vtk(num_array=data, deep=True)
        data.SetName('Spatial Point Data')
        # THIS IS CELL DATA! Add the model data to CELL data:
        pdo.GetPointData().AddArray(data)
        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        # Now set whole output extent
        ext = [0, self.__extent[0]-1, 0,self.__extent[1]-1, 0,self.__extent[2]-1]
        info = outInfo.GetInformationObject(0)
        # Set WHOLE_EXTENT: This is absolutely necessary
        info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), ext, 6)
        return 1


    #### Setters / Getters ####


    def SetExtent(self, nx, ny, nz):
        """Set the extent of the output grid.
        """
        if self.__extent != [nx, ny, nz]:
            self.__extent = [nx, ny, nz]
            self.Modified()

    def SetSpacing(self, dx, dy, dz):
        """Set the spacing for the points along each axial direction.
        """
        if self.__spacing != [dx, dy, dz]:
            self.__spacing = [dx, dy, dz]
            self.Modified()

    def SetOrigin(self, x0, y0, z0):
        """Set the origin of the output grid.
        """
        if self.__origin != [x0, y0, z0]:
            self.__origin = [x0, y0, z0]
            self.Modified()





class CreateEvenRectilinearGrid(AlgorithmBase):
    """This creates a vtkRectilinearGrid where the discretization along a given axis is uniformly distributed.
    """
    __displayname__ = 'Create Even Rectilinear Grid'
    __type__ = 'source'
    def __init__(self,
                 extent=[10, 10, 10],
                 xrng=[-1.0, 1.0],
                 yrng=[-1.0, 1.0],
                 zrng=[-1.0, 1.0]):
        AlgorithmBase.__init__(self,
            nInputPorts=0,
            nOutputPorts=1, outputType='vtkRectilinearGrid')
        self.__extent = extent
        self.__xrange = xrng
        self.__yrange = yrng
        self.__zrange = zrng


    def RequestData(self, request, inInfo, outInfo):
        # Get output of Proxy
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        nx,ny,nz = self.__extent[0]+1, self.__extent[1]+1, self.__extent[2]+1

        xcoords = np.linspace(self.__xrange[0], self.__xrange[1], num=nx)
        ycoords = np.linspace(self.__yrange[0], self.__yrange[1], num=ny)
        zcoords = np.linspace(self.__zrange[0], self.__zrange[1], num=nz)

        # CONVERT TO VTK #
        xcoords = nps.numpy_to_vtk(num_array=xcoords,deep=True)
        ycoords = nps.numpy_to_vtk(num_array=ycoords,deep=True)
        zcoords = nps.numpy_to_vtk(num_array=zcoords,deep=True)

        pdo.SetDimensions(nx,ny,nz)
        pdo.SetXCoordinates(xcoords)
        pdo.SetYCoordinates(ycoords)
        pdo.SetZCoordinates(zcoords)

        data = _makeSpatialCellData(nx-1, ny-1, nz-1)
        data = nps.numpy_to_vtk(num_array=data, deep=True)
        data.SetName('Spatial Data')
        # THIS IS CELL DATA! Add the model data to CELL data:
        pdo.GetCellData().AddArray(data)
        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        # Now set whole output extent
        ext = [0, self.__extent[0], 0,self.__extent[1], 0,self.__extent[2]]
        info = outInfo.GetInformationObject(0)
        # Set WHOLE_EXTENT: This is absolutely necessary
        info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), ext, 6)
        return 1


    #### Setters / Getters ####


    def SetExtent(self, nx, ny, nz):
        """Set the extent of the output grid.
        """
        if self.__extent != [nx, ny, nz]:
            self.__extent = [nx, ny, nz]
            self.Modified()

    def SetXRange(self, start, stop):
        """Set range (min, max) for the grid in the X-direction.
        """
        if self.__xrange != [start, stop]:
            self.__xrange = [start, stop]
            self.Modified()

    def SetYRange(self, start, stop):
        """Set range (min, max) for the grid in the Y-direction
        """
        if self.__yrange != [start, stop]:
            self.__yrange = [start, stop]
            self.Modified()

    def SetZRange(self, start, stop):
        """Set range (min, max) for the grid in the Z-direction
        """
        if self.__zrange != [start, stop]:
            self.__zrange = [start, stop]
            self.Modified()



import unittest
import numpy as np

# VTK imports:
import vtk
from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa
from .. import _helpers

# Functionality to test:
from .reverse_axii import *
from .table2grid import *
from .trans_origin import *

RTOL = 0.000001



###############################################################################

class TestTableToGrid(unittest.TestCase):
    """
    Test the `TableToGrid` filter
    """
    # TODO: This is conceptually difficult to decipher whether or not it is working
    def setUp(self):
        # Create some input tables
        self.table = vtk.vtkTable()
        # Populate the tables
        self.arrs = [None, None, None]
        self.n = 400
        self.titles = ('Array 0', 'Array 1', 'Array 2')
        self.arrs[0] = np.random.random(self.n)
        self.arrs[1] = np.random.random(self.n)
        self.arrs[2] = np.random.random(self.n)
        self.table.AddColumn(_helpers.numToVTK(self.arrs[0], self.titles[0]))
        self.table.AddColumn(_helpers.numToVTK(self.arrs[1], self.titles[1]))
        self.table.AddColumn(_helpers.numToVTK(self.arrs[2], self.titles[2]))
        return

    def check_data_fidelity(self, ido):
        """`TableToGrid`: data fidelity"""
        wido = dsa.WrapDataObject(ido)
        for i in range(len(self.titles)):
            arr = wido.PointData[self.titles[i]]
            self.assertTrue(np.allclose(arr, self.arrs[i], rtol=RTOL))

    def test_simple(self):
        """`TableToGrid`: check simple"""
        # Rearange data to check to match table
        for i in range(len(self.arrs)):
            a = np.reshape(self.arrs[i], (20, 2, 10))
            a = np.swapaxes(a, 0, 2)
            self.arrs[i] = a.flatten()
        # Use filter
        f = TableToGrid()
        f.SetInputDataObject(self.table)
        f.SetExtent(20, 2, 10)
        f.SetSpacing(5, 5, 5)
        f.SetOrigin(3.3, 6.0, 7)
        f.Update()
        ido = f.GetOutput()
        self.check_data_fidelity(ido)


    def test_fortran(self):
        """`TableToGrid`: check fortran order"""
        # Rearange data to check to match table
        for i in range(len(self.arrs)):
            a = np.reshape(self.arrs[i], (20, 2, 10))
            a = np.swapaxes(a, 0, 2)
            self.arrs[i] = a.flatten(order='F')
        # Use filter
        f = TableToGrid()
        f.SetInputDataObject(self.table)
        f.SetExtent(20, 2, 10)
        f.SetSpacing(5, 5, 5)
        f.SetOrigin(3.3, 6.0, 7)
        f.SetOrder('F')
        f.Update()
        ido = f.GetOutput()
        self.check_data_fidelity(ido)

    def test_seplib(self):
        """`TableToGrid`: check seplib"""
        # # Rearange data to check to match table
        # for i in range(len(self.arrs)):
        #     a = np.reshape(self.arrs[i], (10, 2, 20))
        #     a = np.swapaxes(a, 0, 2)
        #     a = np.swapaxes(a, 1, 2)
        #     #a = np.swapaxes(a, 0, 2)
        #     self.arrs[i] = a.flatten()
        # Use filter
        f = TableToGrid()
        f.SetInputDataObject(self.table)
        f.SetExtent(20, 2, 10)
        f.SetSpacing(5, 5, 5)
        f.SetOrigin(3.3, 6.0, 7)
        f.SetSEPlib(True)
        f.Update()
        ido = f.GetOutput()
        # TODO: self.check_data_fidelity(ido)
        return

    def test_swapXY(self):
        """`TableToGrid`: check swapXY"""
        # Rearange data to check to match table
        # for i in range(len(self.arrs)):
        #     a = np.reshape(self.arrs[i], (20, 2, 10))
        #     a = np.swapaxes(a, 0, 2)
        #     a = np.swapaxes(a, 1, 2)
        #     self.arrs[i] = a.flatten()
        # Use filter
        f = TableToGrid()
        f.SetInputDataObject(self.table)
        f.SetExtent(20, 2, 10)
        f.SetSpacing(5, 5, 5)
        f.SetOrigin(3.3, 6.0, 7)
        f.SetSwapXY(True)
        f.Update()
        ido = f.GetOutput()
        # TODO: self.check_data_fidelity(ido)
        return



###############################################################################

class TestReverseImageDataAxii(unittest.TestCase):
    """
    Test the `ReverseImageDataAxii` filter
    """

    def test(self):
        """`ReverseImageDataAxii`: check that it works on point and cell data"""
        # Create input vtkImageData:
        nx, ny, nz = 10, 11, 12
        arr = np.random.random((nz, ny, nx))
        arrCells = np.random.random((nz-1, ny-1, nx-1))
        image = vtk.vtkImageData()
        image.SetDimensions(nx, ny, nz)
        image.SetSpacing(2, 2, 2)
        image.SetOrigin(0, 0, 0)
        data = nps.numpy_to_vtk(num_array=arr.flatten(), deep=True)
        data.SetName('Data')
        cellData = nps.numpy_to_vtk(num_array=arrCells.flatten(), deep=True)
        cellData.SetName('CellData')
        image.GetPointData().AddArray(data)
        image.GetCellData().AddArray(cellData)
        # Now perfrom the reverse for only X:
        f = ReverseImageDataAxii()
        f.SetInputDataObject(image)
        f.SetFlipX(True)
        f.SetFlipY(False)
        f.SetFlipZ(False)
        f.Update()
        ido = f.GetOutput()
        self.assertIsNotNone(ido)
        self.assertEqual(ido.GetNumberOfPoints(), len(arr.flatten()))
        # Check that x-axis was reversed
        wido = dsa.WrapDataObject(ido)
        test = wido.PointData['Data']
        testCells = wido.CellData['CellData']
        self.assertTrue(np.allclose(test, np.flip(arr, axis=2).flatten(), rtol=RTOL))
        self.assertTrue(np.allclose(testCells, np.flip(arrCells, axis=2).flatten(), rtol=RTOL))
        # Now perfrom the reverse for all axii:
        f.SetFlipX(True)
        f.SetFlipY(True)
        f.SetFlipZ(True)
        f.Update()
        ido = f.GetOutput()
        self.assertIsNotNone(ido)
        self.assertEqual(ido.GetNumberOfPoints(), len(arr.flatten()))
        # Check that x-axis was reversed
        wido = dsa.WrapDataObject(ido)
        test = wido.PointData['Data']
        testCells = wido.CellData['CellData']
        tarr = np.flip(arr, axis=0)
        tarr = np.flip(tarr, axis=1)
        tarr = np.flip(tarr, axis=2)
        tarrCells = np.flip(arrCells, axis=0)
        tarrCells = np.flip(tarrCells, axis=1)
        tarrCells = np.flip(tarrCells, axis=2)
        self.assertTrue(np.allclose(test, tarr.flatten(), rtol=RTOL))
        self.assertTrue(np.allclose(testCells, tarrCells.flatten(), rtol=RTOL))


###############################################################################

class TestTranslateGridOrigin(unittest.TestCase):
    """
    Test the `TranslateGridOrigin` filter
    """
    def setUp(self):
        # Create some input tables
        self.idi = vtk.vtkImageData()
        self.idi.SetDimensions(20, 2, 10)
        self.idi.SetSpacing(1, 1, 1)
        self.idi.SetOrigin(100, 100, 100)
        # Populate the tables
        self.n = 400
        self.title = 'Array 0'
        self.arr = np.random.random(self.n)
        self.idi.GetPointData().AddArray(_helpers.numToVTK(self.arr, self.title))
        return

    def test_all(self):
        """`TranslateGridOrigin`: make sure works"""
        f = TranslateGridOrigin()
        f.SetInputDataObject(self.idi)
        f.SetCorner(1)
        f.Update()
        ido = f.GetOutput()
        ox, oy, oz = ido.GetOrigin()
        self.assertEqual(ox, 100-19)
        self.assertEqual(oy, 100)
        self.assertEqual(oz, 100)
        f.SetCorner(2)
        f.Update()
        ido = f.GetOutput()
        ox, oy, oz = ido.GetOrigin()
        self.assertEqual(ox, 100)
        self.assertEqual(oy, 100-1)
        self.assertEqual(oz, 100)
        f.SetCorner(3)
        f.Update()
        ido = f.GetOutput()
        ox, oy, oz = ido.GetOrigin()
        self.assertEqual(ox, 100-19)
        self.assertEqual(oy, 100-1)
        self.assertEqual(oz, 100)
        f.SetCorner(4)
        f.Update()
        ido = f.GetOutput()
        ox, oy, oz = ido.GetOrigin()
        self.assertEqual(ox, 100)
        self.assertEqual(oy, 100)
        self.assertEqual(oz, 100-9)
        f.SetCorner(5)
        f.Update()
        ido = f.GetOutput()
        ox, oy, oz = ido.GetOrigin()
        self.assertEqual(ox, 100-19)
        self.assertEqual(oy, 100)
        self.assertEqual(oz, 100-9)
        f.SetCorner(6)
        f.Update()
        ido = f.GetOutput()
        ox, oy, oz = ido.GetOrigin()
        self.assertEqual(ox, 100)
        self.assertEqual(oy, 100-1)
        self.assertEqual(oz, 100-9)
        f.SetCorner(7)
        f.Update()
        ido = f.GetOutput()
        ox, oy, oz = ido.GetOrigin()
        self.assertEqual(ox, 100-19)
        self.assertEqual(oy, 100-1)
        self.assertEqual(oz, 100-9)


###############################################################################



__all__ = [
    'CreateTensorMesh',
]

import vtk
from vtk.util import numpy_support as nps
import numpy as np
from vtk.numpy_interface import dataset_adapter as dsa
# Import Helpers:
from ..base import AlgorithmBase
#from .. import _helpers

class CreateTensorMesh(AlgorithmBase):
    """This creates a vtkRectilinearGrid where the discretization along a given axis is uniformly distributed.
    """
    __displayname__ = 'Create Tensor Mesh'
    __type__ = 'source'
    def __init__(self, origin=[-350.0, -400.0, 0.0], dataname='Data'):
        AlgorithmBase.__init__(self, nInputPorts=0,
            nOutputPorts=1, outputType='vtkRectilinearGrid')
        self.__origin = origin
        self.__xcells = CreateTensorMesh._ReadCellLine('200 100 50 20*50.0 50 100 200')
        self.__ycells = CreateTensorMesh._ReadCellLine('200 100 50 21*50.0 50 100 200')
        self.__zcells = CreateTensorMesh._ReadCellLine('20*25.0 50 100 200')
        self.__dataName = dataname


    @staticmethod
    def _ReadCellLine(line):
        """Read cell sizes for each line in the UBC mesh line strings
        """
        # OPTIMIZE: work in progress
        # TODO: when optimized, make sure to combine with UBC reader
        line_list = []
        for seg in line.split():
            if '*' in seg:
                sp = seg.split('*')
                seg_arr = np.ones((int(sp[0]),), dtype=float) * float(sp[1])
            else:
                seg_arr = np.array([float(seg)], dtype=float)
            line_list.append(seg_arr)
        return np.concatenate(line_list)


    def GetExtent(self):
        ne,nn,nz = len(self.__xcells), len(self.__ycells), len(self.__zcells)
        return (0,ne, 0,nn, 0,nz)



    def _MakeModel(self, pdo):
        ox,oy,oz = self.__origin[0], self.__origin[1], self.__origin[2]

        # Read the cell sizes
        cx = self.__xcells
        cy = self.__ycells
        cz = self.__zcells

        # Invert the indexing of the vector to start from the bottom.
        cz = cz[::-1]
        # Adjust the reference point to the bottom south west corner
        oz = oz - np.sum(cz)

        # Now generate the coordinates for from cell width and origin
        cox = ox + np.cumsum(cx)
        cox = np.insert(cox,0,ox)
        coy = oy + np.cumsum(cy)
        coy = np.insert(coy,0,oy)
        coz = oz + np.cumsum(cz)
        coz = np.insert(coz,0,oz)

        # Set the dims and coordinates for the output
        ext = self.GetExtent()
        nx,ny,nz = ext[1]+1,ext[3]+1,ext[5]+1
        pdo.SetDimensions(nx,ny,nz)
        # Convert to VTK array for setting coordinates
        pdo.SetXCoordinates(nps.numpy_to_vtk(num_array=cox, deep=True))
        pdo.SetYCoordinates(nps.numpy_to_vtk(num_array=coy, deep=True))
        pdo.SetZCoordinates(nps.numpy_to_vtk(num_array=coz, deep=True))

        return pdo


    def _AddModelData(self, pdo, data):
        ext = self.GetExtent()
        nx,ny,nz = ext[1]+1,ext[3]+1,ext[5]+1
        # ADD DATA to cells
        if data is None:
            data = np.random.rand(nx*ny*nz)
            data = nps.numpy_to_vtk(num_array=data, deep=True)
            data.SetName('Random Data')
        else:
            data = nps.numpy_to_vtk(num_array=data, deep=True)
            data.SetName(dataNm)
        pdo.GetCellData().AddArray(data)
        return pdo


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output data object
        """
        # Get input/output of Proxy
        pdo = self.GetOutputData(outInfo, 0)
        # Perform the task
        self._MakeModel(pdo)
        self._AddModelData(pdo, None) # TODO: add ability to set input data
        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        """Used by pipeline to set output whole extent
        """
        # Now set whole output extent
        extent = self.GetExtent()
        ext = [0, extent[0]-2, 0,extent[1]-2, 0,extent[2]-2]
        info = outInfo.GetInformationObject(0)
        # Set WHOLE_EXTENT: This is absolutely necessary
        info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), ext, 6)
        return 1


    #### Getters / Setters ####


    def SetOrigin(self, x0, y0, z0):
        """Set the origin of the output
        """
        if self.__origin != [x0, y0, z0]:
            self.__origin = [x0, y0, z0]
            self.Modified()

    def SetXCells(self, xcells):
        """Set the spacings for the cells in the X direction

        Args:
            xcells (list or np.array(floats)) : the spacings along the X-axis"""
        if len(xcells) != len(self.__xcells) or not np.allclose(self.__xcells, xcells):
            self.__xcells = xcells
            self.Modified()

    def SetYCells(self, ycells):
        """Set the spacings for the cells in the Y direction

        Args:
            ycells (list or np.array(floats)) : the spacings along the Y-axis"""
        if len(ycells) != len(self.__ycells) or not np.allclose(self.__ycells, ycells):
            self.__ycells = ycells
            self.Modified()

    def SetZCells(self, zcells):
        """Set the spacings for the cells in the Z direction

        Args:
            zcells (list or np.array(floats)): the spacings along the Z-axis"""
        if len(zcells) != len(self.__zcells) or not np.allclose(self.__zcells, zcells):
            self.__zcells = zcells
            self.Modified()

    def SetXCellsStr(self, xcellstr):
        """Set the spacings for the cells in the X direction

        Args:
            xcellstr (str) : the spacings along the X-axis in the UBC style"""
        xcells = CreateTensorMesh._ReadCellLine(xcellstr)
        self.SetXCells(xcells)

    def SetYCellsStr(self, ycellstr):
        """Set the spacings for the cells in the Y direction

        Args:
            ycellstr (str) : the spacings along the Y-axis in the UBC style"""
        ycells = CreateTensorMesh._ReadCellLine(ycellstr)
        self.SetYCells(ycells)

    def SetZCellsStr(self, zcellstr):
        """Set the spacings for the cells in the Z direction

        Args:
            zcellstr (str)  : the spacings along the Z-axis in the UBC style"""
        zcells = CreateTensorMesh._ReadCellLine(zcellstr)
        self.SetZCells(zcells)



paraview_plugin_version = '1.1.10'
# This is module to import. It provides VTKPythonAlgorithmBase, the base class
# for all python-based vtkAlgorithm subclasses in VTK and decorators used to
# 'register' the algorithm with ParaView along with information about UI.
from paraview.util.vtkAlgorithm import *

# Helpers:
from PVGeo import _helpers
# Classes to Decorate
from PVGeo.readers import PackedBinariesReader, MadagascarReader
from PVGeo.readers import DelimitedTextReader, XYZTextReader


###############################################################################


PACKED_EXTS = 'H@ bin rsf rsf@ HH npz'
PACKED_DESC = 'Packed Binaries Reader'

@smproxy.reader(name="PVGeoPackedBinariesReader",
       label="PVGeo: Packed Binaries Reader",
       extensions=PACKED_EXTS,
       file_description=PACKED_DESC)
class PVGeoPackedBinariesReader(PackedBinariesReader):
    def __init__(self):
        PackedBinariesReader.__init__(self)

    #### Seters and Geters ####
    @smproperty.xml(_helpers.getFileReaderXml(PACKED_EXTS, readerDescription=PACKED_DESC))
    def AddFileName(self, fname):
        PackedBinariesReader.AddFileName(self, fname)



    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def SetTimeDelta(self, dt):
        PackedBinariesReader.SetTimeDelta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        """This is critical for registering the timesteps"""
        return PackedBinariesReader.GetTimestepValues(self)

    @smproperty.xml(_helpers.getDropDownXml('Endian','SetEndian',
        ['Native', 'Little-Endian', 'Big-Endian'],
        help='This is the type memory endianness.'))
    def SetEndian(self, endian):
        PackedBinariesReader.SetEndian(self, endian)

    @smproperty.xml(_helpers.getDropDownXml('DataType','SetDataType',
        ['Float64', 'Float32', 'Integer4'],
        help='This is data type to read.'))
    def SetDataType(self, dtype):
        PackedBinariesReader.SetDataType(self, dtype)


    @smproperty.stringvector(name='DataName', default_values='Data')
    def SetDataName(self, dataName):
        PackedBinariesReader.SetDataName(self, dataName)


###############################################################################


MADAGASCAR_EXTS = 'bin H@ rsf rsf@ HH'
MADAGASCAR_DESC = 'Madagascar Single Stream RSF Files'

@smproxy.reader(name="PVGeoMadagascarReader",
       label="PVGeo: Madagascar SSRSF  Reader",
       extensions=MADAGASCAR_EXTS,
       file_description=MADAGASCAR_DESC)
class PVGeoMadagascarReader(MadagascarReader):
    def __init__(self):
        MadagascarReader.__init__(self)

    #### Seters and Geters ####


    @smproperty.xml(_helpers.getFileReaderXml(MADAGASCAR_EXTS, readerDescription=MADAGASCAR_DESC))
    def AddFileName(self, fname):
        MadagascarReader.AddFileName(self, fname)


    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def SetTimeDelta(self, dt):
        MadagascarReader.SetTimeDelta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        """This is critical for registering the timesteps"""
        return MadagascarReader.GetTimestepValues(self)

    @smproperty.xml(_helpers.getDropDownXml('Endian','SetEndian',
        ['Native', 'Little-Endian', 'Big-Endian'],
        help='This is the type memory endianness.'))
    def SetEndian(self, endian):
        MadagascarReader.SetEndian(self, endian)

    @smproperty.xml(_helpers.getDropDownXml('DataType','SetDataType',
        ['Float64', 'Float32', 'Integer4'],
        help='This is data type to read.'))
    def SetDataType(self, dtype):
        MadagascarReader.SetDataType(self, dtype)


    @smproperty.stringvector(name='DataName', default_values='Data')
    def SetDataName(self, dataName):
        MadagascarReader.SetDataName(self, dataName)



###############################################################################


DELIMITED_EXTS = 'dat csv txt text ascii xyz'
DELIMITED_DESC = 'Delimited Text Files'

@smproxy.reader(name="PVGeoDelimitedTextReader",
       label="PVGeo: Delimited Text Reader",
       extensions=DELIMITED_EXTS,
       file_description=DELIMITED_DESC)
class PVGeoDelimitedTextReader(DelimitedTextReader):
    def __init__(self):
        DelimitedTextReader.__init__(self)


    #### Seters and Geters ####
    @smproperty.xml(_helpers.getFileReaderXml(DELIMITED_EXTS, readerDescription=DELIMITED_DESC))
    def AddFileName(self, fname):
        DelimitedTextReader.AddFileName(self, fname)



    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def SetTimeDelta(self, dt):
        DelimitedTextReader.SetTimeDelta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        """This is critical for registering the timesteps"""
        return DelimitedTextReader.GetTimestepValues(self)

    @smproperty.stringvector(name="Delimiter", default_values=" ")
    def SetDelimiter(self, deli):
        DelimitedTextReader.SetDelimiter(self, deli)

    @smproperty.xml(_helpers.getPropertyXml(name='Use Tab Delimiter', command='SetUseTab', default_values=False, help='A boolean to override the Delimiter_Field and use a Tab delimiter.'))
    def SetUseTab(self, flag):
        DelimitedTextReader.SetUseTab(self, flag)

    @smproperty.intvector(name="SkipRows", default_values=0)
    def SetSkipRows(self, skip):
        DelimitedTextReader.SetSkipRows(self, skip)

    @smproperty.stringvector(name="Comments", default_values="#")
    def SetComments(self, identifier):
        DelimitedTextReader.SetComments(self, identifier)

    @smproperty.xml(_helpers.getPropertyXml(name='Has Titles', command='SetHasTitles', default_values=True, help='A boolean for if the delimited file has header titles for the data arrays.'))
    def SetHasTitles(self, flag):
        DelimitedTextReader.SetHasTitles(self, flag)



###############################################################################


XYZ_EXTS = 'xyz text txt dat csv tsv'
XYZ_DESC = 'XYZ Delimited Text Files where header has comma delimiter.'

@smproxy.reader(name="PVGeoXYZTextReader",
       label="PVGeo: XYZ File Reader",
       extensions=XYZ_EXTS,
       file_description=XYZ_DESC)
class PVGeoXYZTextReader(XYZTextReader):
    def __init__(self):
        XYZTextReader.__init__(self)


    #### Seters and Geters ####
    @smproperty.xml(_helpers.getFileReaderXml(XYZ_EXTS, readerDescription=XYZ_DESC))
    def AddFileName(self, fname):
        XYZTextReader.AddFileName(self, fname)



    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def SetTimeDelta(self, dt):
        XYZTextReader.SetTimeDelta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        """This is critical for registering the timesteps"""
        return XYZTextReader.GetTimestepValues(self)

    @smproperty.stringvector(name="Delimiter", default_values=" ")
    def SetDelimiter(self, deli):
        XYZTextReader.SetDelimiter(self, deli)

    @smproperty.xml(_helpers.getPropertyXml(name='Use Tab Delimiter', command='SetUseTab', default_values=False, help='A boolean to override the Delimiter_Field and use a Tab delimiter.'))
    def SetUseTab(self, flag):
        XYZTextReader.SetUseTab(self, flag)

    @smproperty.intvector(name="SkipRows", default_values=0)
    def SetSkipRows(self, skip):
        XYZTextReader.SetSkipRows(self, skip)

    @smproperty.stringvector(name="Comments", default_values="#")
    def SetComments(self, identifier):
        XYZTextReader.SetComments(self, identifier)



"""
These are helpers specifically for the file readers for private use only.
@author: Bane Sullivan
"""
__all__ = [
    'getVTKtype',
    'placeArrInTable',
    'getdTypes',
    'cleanDataNm'
]

import numpy as np
from vtk.util import numpy_support as nps
import vtk
import os
from . import errors as _helpers

def getVTKtype(typ):
    """This looks up the VTK type for a give python data type.

    Return:
        int : the integer type id specified in vtkType.h
    """
    typ = nps.get_vtk_array_type(typ)
    if typ is 3:
        return 13
    return typ

def converStringArray(arr):
    """A helper to convert a numpy array of strings to a vtkStringArray

    Return:
        vtkStringArray : the converted array
    """
    vtkarr = vtk.vtkStringArray()
    for val in arr:
        vtkarr.InsertNextValue(val)
    return vtkarr

def ConvertArray(arr):
    """A helper to convert a numpy array to a vtkDataArray

    Return:
        vtkDataArray : the converted array

    Note:
        this converts the data array but does not set a name. The name must be set for this data array to be added to a vtkDataSet ``array.SetName('Data')``
    """
    typ = getVTKtype(arr.dtype)
    if typ is 13:
        VTK_data = converStringArray(arr)
    else:
        VTK_data = nps.numpy_to_vtk(num_array=arr, deep=True, array_type=typ)
    return VTK_data


def placeArrInTable(ndarr, titles, pdo):
    """Takes a 1D/2D numpy array and makes a vtkTable of it

    Args:
        ndarr (numpy.ndarray) : The 1D/2D array to be converted to a table
        titles (list or tuple): The titles for the arrays in the table. Must have same number of elements as columns in input ndarray
        pdo (vtkTable) : The output data object pointer

    Return:
        vtkTable : returns the same input pdo table
    """
    # Put columns into table
    if len(np.shape(ndarr)) > 2:
        raise _helpers.PVGeoError('Input np.ndarray must be 1D or 2D to be converted to vtkTable.')
    if len(np.shape(ndarr)) == 1:
        # First check if it is an array full of tuples (varying type)
        if isinstance(ndarr[0], (tuple, np.void)):
            for i in range(len(titles)):
                placeArrInTable(ndarr['f%d' % i], [titles[i]], pdo)
            return pdo
        # Otherwise it is just a 1D array which needs to be 2D
        else:
            ndarr = np.reshape(ndarr, (-1, 1))
    cols = np.shape(ndarr)[1]

    for i in range(cols):
        VTK_data = ConvertArray(ndarr[:,i])
        VTK_data.SetName(titles[i])
        pdo.AddColumn(VTK_data)
    return pdo



def getdTypes(dtype='', endian=None):
    """This converts char dtypes and an endian to a numpy and VTK data type.

    Return:
        tuple (numpy.dtype, int) : the numpy data type and the integer type id specified in vtkType.h for VTK data types
    """
    # If native `@` was chosen then do not pass an endian
    if endian is '@':
        #print('WARNING: Native endianness no longer supported for packed binary reader. Please chose `>` or `<`. This defaults to big `>`.')
        endian = ''
    # No endian specified:
    elif endian is None:
        endian = ''
    # Get numpy and VTK data types and return them both
    if dtype is 'd':
        vtktype = vtk.VTK_DOUBLE
    elif dtype is 'f':
        vtktype = vtk.VTK_FLOAT
    elif dtype is 'i':
        vtktype = vtk.VTK_INT
    else:
        raise _helpers.PVGeoError('dtype \'%s\' unknown:' % dtype)
    # Return data types
    dtype = np.dtype('%s%s' % (endian, dtype))
    return dtype, vtktype


def cleanDataNm(dataNm, FileName):
    """A helper to clean a FileName to make a useful data array name"""
    if dataNm is None or dataNm == '':
        dataNm = os.path.splitext(os.path.basename(FileName))[0]
    return dataNm



__all__ = ['camera']

import numpy as np
import os
import pickle
from paraview.simple import RenderAllViews, GetActiveCamera, WriteImage

class camera:
    """An object to store a single camera location/view. You can make a list/dict of these objects to save interesting views for your project. This object saves just a few parameters about the camera so that it can easily be reconstructed.
    """
    __displayname__ = 'Camera'
    __type__ = 'macro'
    def __init__(self,cam=None):
        """
        @params:
        cam : vtkRenderingOpenGL2Python.vtkOpenGLCamera : optional : The camera you wish to update this object to. Totally optional
        """
        if cam is None:
            # This allows use to dynamicly select cameras
            cam = GetActiveCamera()
        self.orientation = cam.GetOrientation()
        self.position = cam.GetPosition()
        self.focus = cam.GetFocalPoint()
        self.viewup = cam.GetViewUp()

    # Variable access for internal use
    def _getOrientation(self):
        return self.orientation

    def _getPosition(self):
        return self.position

    def _getFocalPoint(self):
        return self.focus

    def _getViewUp(self):
        return self.viewup

    # Use new location
    def update(self, cam=None):
        """Updates the camera location to that which is in the currently activated view unless a vtkOpenGLCamera is specified.

        Args:
            cam (vtkRenderingOpenGL2Python.vtkOpenGLCamera) : The camera you wish to update this object to. Totally optional
        """
        if cam is None:
            # This allows use to dynamicly select cameras
            cam = GetActiveCamera()
        self.orientation = cam.GetOrientation()
        self.position = cam.GetPosition()
        self.focus = cam.GetFocalPoint()
        self.viewup = cam.GetViewUp()

    # Change the camera view
    def view(self, cam=None):
        """Use this method to update the camera to the saved location

        Args:
            cam (vtkRenderingOpenGL2Python.vtkOpenGLCamera) : The camera you wish to view/update in the current render view
        """
        if cam is None:
            # This allows use to dynamicly select cameras
            cam = GetActiveCamera()
        orientation = self._getOrientation()
        position = self._getPosition()
        focus = self._getFocalPoint()
        viewup = self._getViewUp()

        # set the camera position and orientation
        cam.SetPosition(position)
        cam.SetViewUp(viewup)
        cam.SetFocalPoint(focus)
        RenderAllViews()

    # Save Screenshot of single view
    def screenShot(self, cam=None, path=os.path.expanduser('~'), basenm='view'):
        """Save a screenshot of a single camera view

        Args:
            cam (vtkRenderingOpenGL2Python.vtkOpenGLCamera) : The camera you wish to view then save a screenshot
            path (str) : The directory you wish to save the screenshot. Defaults to user home directory
            basenm (str) : The file basename for the screenshot
        """
        if cam is None:
            # This allows use to dynamicly select cameras
            cam = GetActiveCamera()
        os.chdir(path)
        self.view(cam=cam)
        WriteImage("%s.png" % (basenm))


    # Static Methods for structures containt cameras

    @staticmethod
    def saveViews(lib, filename='views', path=os.path.expanduser('~')):
        """Save a serialized dictionaty/list/whatever of views out to a file. Dafault saves to user's home directory

        Args:
            lib (dict or list) : some iterable object containg multiple `camera` objects
            filename (str) : The file basename for the serialized file
            path (str) : The directory you wish to save the views. Defaults to user home directory
        """
        ext = '.camera'
        os.chdir(path)
        f = open(filename + ext, 'wb')
        pickle.dump(lib, f, pickle.HIGHEST_PROTOCOL)
        f.close()

    @staticmethod
    def loadViews(filename='views.camera', path=os.path.expanduser('~')):
        """Load a file containg a serialized camera objects. Dafault loads from home directory if relative path

        Args:
            filename (str) : The file basename for the serialized file (defualt is default for output def)
            path (str): The directory from which you wish to load the views. Defaults to user home directory for relative paths.
        """
        os.chdir(path)
        with open(filename, 'rb') as f:
            return pickle.load(f)

    @staticmethod
    def screenShotViews(views, cam=None, path=os.path.expanduser('~'), basenm='view'):
        """Save screenshots of many views/cameras

        Args:
            view d(ict or list) : some iterable object containg multiple `camera` objects
            cam (vtkRenderingOpenGL2Python.vtkOpenGLCamera) : The camera you wish to view then save a screenshot
            path (str): The directory you wish to save the screenshot. Defaults to user home directory
            basenm (str): The file basename for the screenshot
        """
        if cam is None:
            # This allows use to dynamicly select cameras
            cam = GetActiveCamera()

        def _iter(obj):
            return obj if isinstance(obj, dict) else xrange(len(obj))

        os.chdir(path)
        for v in _iter(views):
            views[v].view(cam=cam)
            WriteImage("%s_%s.png" % (basenm, v))



__all__ = [
    'ArrayMath',
    'NormalizeArray',
    'AddCellConnToPoints',
    'PointsToTube',
]

import vtk
from vtk.util import numpy_support as nps
import numpy as np
from vtk.numpy_interface import dataset_adapter as dsa
from datetime import datetime
# Import Helpers:
from ..base import FilterBase, FilterPreserveTypeBase
from .. import _helpers
# NOTE: internal import - from scipy.spatial import cKDTree



###############################################################################

#---- ArrayMath ----#
class ArrayMath(FilterPreserveTypeBase):
    """This filter allows the user to select two input data arrays on which to perfrom math operations. The input arrays are used in their order of selection for the operations.

    **Available Math Operations:**

    - `add`: This adds the two data arrays together
    - `subtract`: This subtracts input array 2 from input array 1
    - `multiply`: Multiplies the two data arrays together
    - `divide`: Divide input array 1 by input array 2 (arr1/arr2)
    - `correlate`: Use `np.correlate(arr1, arr2, mode='same')`
    """
    __displayname__ = 'Array Math'
    __type__ = 'filter'
    def __init__(self, **kwargs):
        FilterPreserveTypeBase.__init__(self)
        # Parameters:
        self.__multiplier = kwargs.get('multiplier', 1.0)
        self.__newName = kwargs.get('newName', 'Mathed Up')
        self.__inputArray1 = [None, None]
        self.__inputArray2 = [None, None]
        # Convert operation to callable method
        op = kwargs.get('operation', 'add')
        if isinstance(op, (str, int)):
            op = self.GetOperation(op)
        self.__operation = op


    @staticmethod
    def _correlate(arr1, arr2):
        """Use ``np.correlate()`` on ``mode='same'`` on two selected arrays from one input.
        """
        return np.correlate(arr1, arr2, mode='same')

    @staticmethod
    def _multiply(arr1, arr2):
        return arr1*arr2

    @staticmethod
    def _divide(arr1, arr2):
        return arr1/arr2

    @staticmethod
    def _add(arr1, arr2):
        return arr1+arr2

    @staticmethod
    def _subtract(arr1, arr2):
        return arr1-arr2

    @staticmethod
    def GetOperations():
        """Returns the math operation methods as callable objects in a dictionary
        """
        ops = dict(
            add=ArrayMath._add,
            subtract=ArrayMath._subtract,
            multiply=ArrayMath._multiply,
            divide=ArrayMath._divide,
            correlate=ArrayMath._correlate,
        )
        return ops

    @staticmethod
    def GetOperationNames():
        """Gets a list of the math operation keys

        Return:
            list(str): the keys for getting the math operations
        """
        ops = ArrayMath.GetOperations()
        return list(ops.keys())

    @staticmethod
    def GetOperation(idx):
        """Gets a math operation based on an index in the keys

        Return:
            callable: the math operation method
        """
        if isinstance(idx, str):
            return ArrayMath.GetOperations()[idx]
        n = ArrayMath.GetOperationNames()[idx]
        return ArrayMath.GetOperations()[n]


    def _MathUp(self, pdi, pdo):
        """Make sure to pass array names and integer associated fields.
        Use helpers to get these properties.
        """
        if pdo is None:
            # TODO: test this
            pdo = pdi.DeepCopy()
        # Get the input arrays
        field1, name1 = self.__inputArray1[0], self.__inputArray1[1]
        field2, name2 = self.__inputArray2[0], self.__inputArray2[1]
        wpdi = dsa.WrapDataObject(pdi)
        arr1 = _helpers.getArray(wpdi, field1, name1)
        arr2 = _helpers.getArray(wpdi, field2, name2)
        # Perform Math Operation
        carr = self.__operation(arr1, arr2)
        # Apply the multiplier
        carr *= self.__multiplier
        # Convert to a VTK array
        c = nps.numpy_to_vtk(num_array=carr,deep=True)
        # If no name given for data by user, use operator name
        newName = self.__newName
        if newName == '':
            newName = 'Mathed Up'
        c.SetName(newName)
        # Build output
        pdo.DeepCopy(pdi)
        pdo = _helpers.addArray(pdo, field1, c)
        return pdo


    #### Algorithm Methods ####


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to perfrom operation and generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._MathUp(pdi, pdo)
        return 1


    #### Seters and Geters ####


    def _SetInputArray1(self, field, name):
        if self.__inputArray1[0] != field:
            self.__inputArray1[0] = field
            self.Modified()
        if self.__inputArray1[1] != name:
            self.__inputArray1[1] = name
            self.Modified()

    def _SetInputArray2(self, field, name):
        if self.__inputArray2[0] != field:
            self.__inputArray2[0] = field
            self.Modified()
        if self.__inputArray2[1] != name:
            self.__inputArray2[1] = name
            self.Modified()

    def SetInputArrayToProcess(self, idx, port, connection, field, name):
        """Used by pipeline/paraview GUI wrappings to set the input arrays
        """
        if idx == 0:
            self._SetInputArray1(field, name)
        elif idx == 1:
            self._SetInputArray2(field, name)
        else:
            raise _helpers.PVGeoError('SetInputArrayToProcess() do not know how to handle idx: %d' % idx)
        return 1

    def SetMultiplier(self, val):
        """This is a static shifter/scale factor across the array after normalization.
        """
        if self.__multiplier != val:
            self.__multiplier = val
            self.Modified()

    def GetMultiplier(self):
        """Return the set multiplier/scalar
        """
        return self.__multiplier

    def SetNewArrayName(self, name):
        """Give the new array a meaningful name.
        """
        if self.__newName != name:
            self.__newName = name
            self.Modified()

    def GetNewArrayName(self):
        return self.__newName

    def SetOperation(self, op):
        """Set the math operation to perform

        Args:
            op (str, int, or callable): The operation as a string key, int index, or callable method

        Note:
            This can accept a callable method to set a custom operation as long as its signature is: ``<callable>(arr1, arr2)``
        """
        if isinstance(op, str):
            op = ArrayMath.GetOperations()[op]
        elif isinstance(op, int):
            op = ArrayMath.GetOperation(op)
        if self.__operation != op:
            self.__operation = op
            self.Modified()



###############################################################################

#---- Normalizations ----#

class NormalizeArray(FilterPreserveTypeBase):
    """This filter allows the user to select an array from the input data set to be normalized. The filter will append another array to that data set for the output. The user can specify how they want to rename the array, can choose a multiplier, and can choose from several types of common normalizations (more functionality added as requested).

    **Normalization Types:**

    - `feature_scale`: Feature Scale
    - `standard_score`: tandard Score
    - `log10`: Natural Log
    - `natural_log`: Log Base 10
    - `just_multiply`: Only Multiply by Multiplier
    """
    __displayname__ = 'Normalize Array'
    __type__ = 'filter'
    def __init__(self, **kwargs):
        FilterPreserveTypeBase.__init__(self)
        # Parameters:
        self.__multiplier = kwargs.get('multiplier', 1.0)
        self.__newName = kwargs.get('newName', 'Normalized')
        self.__absolute = kwargs.get('absolute', False)
        self.__inputArray = [None, None]
        # Convert operation to callable method
        op = kwargs.get('normalization', 'feature_scale')
        if isinstance(op, (str, int)):
            op = self.GetNormalization(op)
        self.__normalization = op


    #### Array normalization methods ####


    @staticmethod
    def _passArray(arr):
        return np.array(arr)

    @staticmethod
    def _featureScale(arr):
        # TODO: implement ability to use custom range
        # if rng is not None:
        #     mi = rng[0]
        #     ma = rng[1]
        # else:
        mi = np.min(arr)
        ma = np.max(arr)
        return (arr - mi) / (ma - mi)

    @staticmethod
    def _standardScore(arr):
        return (arr - np.mean(arr)) / (np.std(arr))

    @staticmethod
    def _log10(arr):
        return np.log10(arr)

    @staticmethod
    def _logNat(arr):
        return np.log(arr)

    @staticmethod
    def GetNormalizations():
        """All Available normalizations

        Return:
            dict: dictionary of callable methods for normalizing an array
        """
        ops = dict(
            feature_scale=NormalizeArray._featureScale,
            standard_score=NormalizeArray._standardScore,
            log10=NormalizeArray._log10,
            natural_log=NormalizeArray._logNat,
            just_multiply=NormalizeArray._passArray,
        )
        return ops

    @staticmethod
    def GetNormalizationNames():
        """Gets a list of the normalization keys

        Return:
            list(str): the keys for getting the normalizations
        """
        ops = NormalizeArray.GetNormalizations()
        return list(ops.keys())

    @staticmethod
    def GetNormalization(idx):
        """Gets a normalization based on an index in the keys

        Return:
            callable: the normalization method
        """
        if isinstance(idx, str):
            return NormalizeArray.GetNormalizations()[idx]
        n = NormalizeArray.GetNormalizationNames()[idx]
        return NormalizeArray.GetNormalizations()[n]

    @staticmethod
    def GetArrayRange(pdi, field, name):
        """Returns a tuple of the range for a ``vtkDataArray`` on a ``vtkDataObject``
        """
        wpdi = dsa.WrapDataObject(pdi)
        arr = _helpers.getArray(wpdi, field, name)
        arr = np.array(arr)
        return (np.min(arr), np.max(arr))


    def _Normalize(self, pdi, pdo):
        """Perform normalize on a data array for any given VTK data object.
        """
        # Get inout array
        field, name = self.__inputArray[0], self.__inputArray[1]
        #self.__range = NormalizeArray.GetArrayRange(pdi, field, name)
        wpdi = dsa.WrapDataObject(pdi)
        arr = _helpers.getArray(wpdi, field, name)
        arr = np.array(arr, dtype=float)
        # Take absolute value?
        if self.__absolute:
            arr = np.abs(arr)
        # Perform normalization scheme
        arr = self.__normalization(arr)
        # Apply the multiplier
        arr *= self.__multiplier
        # Convert to VTK array
        c = nps.numpy_to_vtk(num_array=arr,deep=True)
        # If no name given for data by user, use operator name
        newName = self.__newName
        if newName == '':
            newName = 'Normalized ' + name
        c.SetName(newName)
        # Build output
        pdo.DeepCopy(pdi)
        pdo = _helpers.addArray(pdo, field, c)
        return pdo

    #### Algorithm Methods ####


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._Normalize(pdi, pdo)
        return 1

    #### Seters and Geters ####


    def SetInputArrayToProcess(self, idx, port, connection, field, name):
        """Used by pipeline/paraview GUI wrappings to set the input arrays
        """
        if self.__inputArray[0] != field:
            self.__inputArray[0] = field
            self.Modified()
        if self.__inputArray[1] != name:
            self.__inputArray[1] = name
            self.Modified()
        return 1

    def SetMultiplier(self, val):
        """This is a static shifter/scale factor across the array after normalization.
        """
        if self.__multiplier != val:
            self.__multiplier = val
            self.Modified()


    def GetMultiplier(self):
        """Return the set multiplier/scalar
        """
        return self.__multiplier


    def SetNewArrayName(self, name):
        """Give the new array a meaningful name.
        """
        if self.__newName != name:
            self.__newName = name
            self.Modified()


    def GetNewArrayName(self):
        return self.__newName

    def SetTakeAbsoluteValue(self, flag):
        """This will take the absolute value of the array before normalization.
        """
        if self.__absolute != flag:
            self.__absolute = flag
            self.Modified()

    def SetNormalization(self, norm):
        """Set the normalization operation to perform

        Args:
            norm (str, int, or callable): The operation as a string key, int index, or callable method

        Note:
            This can accept a callable method to set a custom operation as long as its signature is: ``<callable>(arr)``
        """
        if isinstance(norm, str):
            norm = NormalizeArray.GetNormalizations()[norm]
        elif isinstance(norm, int):
            norm = NormalizeArray.GetNormalization(norm)
        if self.__normalization != norm:
            self.__normalization = norm
            self.Modified()



###############################################################################
#---- Cell Connectivity ----#

class AddCellConnToPoints(FilterBase):
    """This filter will add linear cell connectivity between scattered points. You have the option to add ``VTK_Line`` or ``VTK_PolyLine`` connectivity. ``VTK_Line`` connectivity makes a straight line between the points in order (either in the order by index or using a nearest neighbor calculation). The ``VTK_PolyLine`` adds a poly line connectivity between all points as one spline (either in the order by index or using a nearest neighbor calculation). Type map is specified in `vtkCellType.h`.

    **Cell Connectivity Types:**

    - 4: Poly Line
    - 3: Line
    """
    __displayname__ = 'Add Cell Connectivity to Points'
    __type__ = 'filter'
    def __init__(self, **kwargs):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkPolyData',
            nOutputPorts=1, outputType='vtkPolyData')
        # Parameters
        self.__cellType = vtk.VTK_POLY_LINE
        self.__usenbr = kwargs.get('nearestNbr', False)


    def _ConnectCells(self, pdi, pdo, logTime=False):
        """Internal helper to perfrom the connection
        """
        # NOTE: Type map is specified in vtkCellType.h
        cellType = self.__cellType
        nrNbr = self.__usenbr

        if logTime:
            startTime = datetime.now()

        # Get the Points over the NumPy interface
        wpdi = dsa.WrapDataObject(pdi) # NumPy wrapped input
        points = np.array(wpdi.Points) # New NumPy array of poins so we dont destroy input

        def _makePolyCell(ptsi):
            cell = vtk.vtkPolyLine()
            cell.GetPointIds().SetNumberOfIds(len(ptsi))
            for i in ptsi:
                cell.GetPointIds().SetId(i, ptsi[i])
            return cell

        def _makeLineCell(ptsi):
            if len(ptsi) != 2:
                raise _helpers.PVGeoError('_makeLineCell() only handles two points')
            aLine = vtk.vtkLine()
            aLine.GetPointIds().SetId(0, ptsi[0])
            aLine.GetPointIds().SetId(1, ptsi[1])
            return aLine


        cells = vtk.vtkCellArray()
        numPoints = pdi.GetNumberOfPoints()
        if nrNbr:
            from scipy.spatial import cKDTree
            # VTK_Line
            if cellType == vtk.VTK_LINE:
                tree = cKDTree(points)
                ind = tree.query([0.0,0.0,0.0], k=numPoints)[1]
                for i in range(len(ind)-1):
                    # Get indices of k nearest points
                    ptsi = [ind[i], ind[i+1]]
                    cell = _makeLineCell(ptsi)
                    cells.InsertNextCell(cell)
                    points = np.delete(points, 0, 0) # Deletes first row
            # VTK_PolyLine
            elif cellType == vtk.VTK_POLY_LINE:
                tree = cKDTree(points)
                dist, ptsi = tree.query([0.0,0.0,0.0], k=numPoints)
                cell = _makePolyCell(ptsi)
                cells.InsertNextCell(cell)
            else:
                raise _helpers.PVGeoError('Cell Type %d not ye implemented.' % cellType)
        else:
            # VTK_PolyLine
            if cellType == vtk.VTK_POLY_LINE:
                ptsi = [i for i in range(numPoints)]
                cell = _makePolyCell(ptsi)
                cells.InsertNextCell(cell)
            # VTK_Line
            elif cellType == vtk.VTK_LINE:
                for i in range(0, numPoints-1):
                    ptsi = [i, i+1]
                    cell = _makeLineCell(ptsi)
                    cells.InsertNextCell(cell)
            else:
                raise _helpers.PVGeoError('Cell Type %d not ye implemented.' % cellType)

        if logTime:
            print((datetime.now() - startTime))
        # Now add points and cells to output
        pdo.SetPoints(pdi.GetPoints())
        pdo.SetLines(cells)
        # copy point data
        _helpers.copyArraysToPointData(pdi, pdo, 0) # 0 is point data
        return pdo

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output data object
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._ConnectCells(pdi, pdo)
        return 1


    #### Seters and Geters ####


    def SetCellType(self, cellType):
        """Set the cell typ by the integer id as specified in `vtkCellType.h`
        """
        if cellType != self.__cellType:
            self.__cellType = cellType
            self.Modified()

    def SetUseNearestNbr(self, flag):
        """Set a flag on whether to use SciPy's ``cKDTree`` nearest neighbor algorithms to sort the points to before adding linear connectivity.
        """
        if flag != self.__usenbr:
            self.__usenbr = flag
            self.Modified()




###############################################################################


class PointsToTube(AddCellConnToPoints):
    """Takes points from a vtkPolyData object and constructs a line of those points then builds a polygonal tube around that line with some specified radius and number of sides.
    """
    __displayname__ = 'Points to Tube'
    __type__ = 'filter'
    def __init__(self, **kwargs):
        AddCellConnToPoints.__init__(self, **kwargs)
        # Additional Parameters
        # NOTE: CellType should remain vtk.VTK_POLY_LINE (4) connection
        self.__numSides = 20
        self.__radius = 10.0


    def _ConnectCells(self, pdi, pdo, logTime=False):
        """This uses the parent's ``_ConnectCells()`` to build a tub around
        """
        AddCellConnToPoints._ConnectCells(self, pdi, pdo, logTime=logTime)
        tube = vtk.vtkTubeFilter()
        tube.SetInputData(pdo)
        tube.SetRadius(self.__radius)
        tube.SetNumberOfSides(self.__numSides)
        tube.Update()
        pdo.ShallowCopy(tube.GetOutput())
        return pdo


    #### Seters and Geters ####

    def SetRadius(self, radius):
        """Set the radius of the tube
        """
        if self.__radius != radius:
            self.__radius = radius
            self.Modified()

    def SetNumberOfSides(self, num):
        """Set the number of sides (resolution) for the tube
        """
        if self.__numSides != num:
            self.__numSides = num
            self.Modified()



###############################################################################



__all__ = ['customAxisTicks']

def customAxisTicks(rng, axis=0, uniform=False):
    """Use to set custm axis ticks in the render view
    """
    from paraview.simple import GetActiveViewOrCreate, RenderAllViews
    # note that third parameter is the step size
    # get the active view
    rv = GetActiveViewOrCreate('RenderView')
    if axis is 0 or uniform:
        rv.AxesGrid.XAxisUseCustomLabels = 1
        rv.AxesGrid.XAxisLabels = rng
    if axis is 1 or uniform:
        rv.AxesGrid.YAxisUseCustomLabels = 1
        rv.AxesGrid.YAxisLabels = rng
    if axis is 2 or uniform:
        rv.AxesGrid.ZAxisUseCustomLabels = 1
        rv.AxesGrid.ZAxisLabels = rng
    RenderAllViews()
    return None

customAxisTicks.__displayname__ = 'Custom Axis Ticks'
customAxisTicks.__type__ = 'macro'



__all__ = [
    'TensorMeshReader',
    'TensorMeshAppender',
]

import numpy as np
from vtk.util import numpy_support as nps
import vtk
import os

from ..base import AlgorithmBase
from .two_file_base import ubcMeshReaderBase, ModelAppenderBase
from .. import _helpers



class TensorMeshReader(ubcMeshReaderBase):
    """UBC Mesh 2D/3D models are defined using a 2-file format. The "mesh" file
    describes how the data is discretized. The "model" file lists the physical
    property values for all cells in a mesh. A model file is meaningless without
    an associated mesh file. The reader will automatically detect if the mesh is
    2D or 3D and read the remainder of the data with that dimensionality
    assumption. If the mesh file is 2D, then then model file must also be in the
    2D format (same for 3D).

    Note:
        Model File is optional. Reader will still construct ``vtkRectilinearGrid`` safely.
    """
    __displayname__ = 'UBC Tensor Mesh Reader'
    __type__ = 'reader'
    def __init__(self, nOutputPorts=1, outputType='vtkRectilinearGrid', **kwargs):
        ubcMeshReaderBase.__init__(self,
            nOutputPorts=nOutputPorts, outputType=outputType, **kwargs)

        self.__mesh = vtk.vtkRectilinearGrid()
        self.__models = []


    @staticmethod
    def PlaceModelOnMesh(mesh, model, dataNm='Data'):
        """Places model data onto a mesh. This is for the UBC Grid data reaers
        to associate model data with the mesh grid.

        Args:
            mesh (vtkRectilinearGrid): The ``vtkRectilinearGrid`` that is the mesh to place the model data upon.
            model (np.array): A NumPy float array that holds all of the data to place inside of the mesh's cells.
            dataNm (str) : The name of the model data array once placed on the ``vtkRectilinearGrid``.

        Return:
            vtkRectilinearGrid : Returns the input ``vtkRectilinearGrid`` with model data appended.
        """
        if type(model) is dict:
            for key in model.keys():
                TensorMeshReader.PlaceModelOnMesh(mesh, model[key], dataNm=key)
            return mesh

        # model.GetNumberOfValues() if model is vtkDataArray
        # Make sure this model file fits the dimensions of the mesh
        ext = mesh.GetExtent()
        n1,n2,n3 = ext[1],ext[3],ext[5]
        if (n1*n2*n3 < len(model)):
            raise _helpers.PVGeoError('Model `%s` has more data than the given mesh has cells to hold.' % dataNm)
        elif (n1*n2*n3 > len(model)):
            raise _helpers.PVGeoError('Model `%s` does not have enough data to fill the given mesh\'s cells.' % dataNm)

        # Swap axes because VTK structures the coordinates a bit differently
        #-  This is absolutely crucial!
        #-  Do not play with unless you know what you are doing!
        model = np.reshape(model, (n1,n2,n3))
        model = np.swapaxes(model,0,1)
        model = np.swapaxes(model,0,2)
        # Now reverse Z axis
        model = model[::-1,:,:] # Note it is in Fortran ordering
        model = model.flatten()

        # Convert data to VTK data structure and append to output
        c = nps.numpy_to_vtk(num_array=model,deep=True)
        c.SetName(dataNm)
        # THIS IS CELL DATA! Add the model data to CELL data:
        mesh.GetCellData().AddArray(c)
        return mesh

    #------------------------------------------------------------------#
    #----------------------     UBC MESH 2D    ------------------------#
    #------------------------------------------------------------------#

    @staticmethod
    def ubcMesh2D(FileName, output):
        """This method reads a UBC 2D Mesh file and builds an empty ``vtkRectilinearGrid``
        for data to be inserted into. `Format Specs`_.

        .. _Format Specs: http://giftoolscookbook.readthedocs.io/en/latest/content/fileFormats/mesh2Dfile.html/>

        Args:
            FileName (str) : The mesh filename as an absolute path for the input mesh file in UBC 3D Mesh Format.
            output (vtkRectilinearGrid) : The output data object

        Return:
            vtkRectilinearGrid : a ``vtkRectilinearGrid`` generated from the UBC 3D Mesh grid. Mesh is defined by the input mesh file. No data attributes here, simply an empty mesh. Use the ``PlaceModelOnMesh()`` method to associate with model data.
        """
        # Read in data from file
        xpts, xdisc, zpts, zdisc = ubcMeshReaderBase._ubcMesh2D_part(FileName)

        nx = np.sum(np.array(xdisc,dtype=int))+1
        nz = np.sum(np.array(zdisc,dtype=int))+1

        # Now generate the vtkRectilinear Grid
        def _genCoords(pts, disc, z=False):
            c = [float(pts[0])]
            for i in range(len(pts)-1):
                start = float(pts[i])
                stop = float(pts[i+1])
                num = int(disc[i])
                w = (stop-start)/num

                for j in range(1,num):
                    c.append(start + (j)*w)
                c.append(stop)
            c = np.array(c,dtype=float)
            if z:
                c = -c[::-1]
            return nps.numpy_to_vtk(num_array=c,deep=True)

        xcoords = _genCoords(xpts, xdisc)
        zcoords = _genCoords(zpts, zdisc, z=True)
        ycoords = nps.numpy_to_vtk(num_array=np.zeros(1),deep=True)

        output.SetDimensions(nx,2,nz) # note this subtracts 1
        output.SetXCoordinates(xcoords)
        output.SetYCoordinates(ycoords)
        output.SetZCoordinates(zcoords)

        return output

    @staticmethod
    def ubcModel2D(FileName):
        """Reads a 2D model file and returns a 1D NumPy float array. Use the ``PlaceModelOnMesh()`` method to associate with a grid.

        Args:
            FileName (str) : The model filename as an absolute path for the input model file in UBCMesh Model Format. Also accepts a list of string file names.

        Return:
            np.array : a NumPy float array that holds the model data read from the file. Use the ``PlaceModelOnMesh()`` method to associate with a grid. If a list of file names is given then it will return a dictionary of NumPy float array with keys as the basenames of the files.
        """
        if type(FileName) is list:
            out = {}
            for f in FileName:
                out[os.path.basename(f)] = TensorMeshReader.ubcModel2D(f)
            return out

        fileLines = np.genfromtxt(FileName, dtype=str, delimiter='\n', comments='!')
        dim = np.array(fileLines[0].split(), dtype=int)
        data = np.genfromtxt((line.encode('utf8') for line in fileLines[1::]), dtype=np.float)
        if np.shape(data)[0] != dim[1] and np.shape(data)[1] != dim[0]:
            raise _helpers.PVGeoError('Mode file `%s` improperly formatted.' % FileName)
        return data.flatten(order='F')


    def __ubcMeshData2D(self, FileName_Mesh, FileName_Models, output):
        """Helper method to read a 2D mesh
        """
        # Construct/read the mesh
        if self.NeedToReadMesh():
            TensorMeshReader.ubcMesh2D(FileName_Mesh, self.__mesh)
            self.NeedToReadMesh(flag=False)
        output.DeepCopy(self.__mesh)
        if self.NeedToReadModels() and self.ThisHasModels():
            self.__models = []
            for f in FileName_Models:
                # Read the model data
                self.__models.append(TensorMeshReader.ubcModel2D(f))
            self.NeedToReadModels(flag=False)
        return output


    #------------------------------------------------------------------#
    #----------------------     UBC MESH 3D    ------------------------#
    #------------------------------------------------------------------#

    @staticmethod
    def ubcMesh3D(FileName, output):
        """This method reads a UBC 3D Mesh file and builds an empty
        ``vtkRectilinearGrid`` for data to be inserted into.

        Args:
            FileName (str) : The mesh filename as an absolute path for the input mesh file in UBC 3D Mesh Format.
            output (vtkRectilinearGrid) : The output data object

        Return:
            vtkRectilinearGrid : a ``vtkRectilinearGrid`` generated from the UBC 3D Mesh grid. Mesh is defined by the input mesh file. No data attributes here, simply an empty mesh. Use the ``PlaceModelOnMesh()`` method to associate with model data.
        """

        #--- Read in the mesh ---#
        fileLines = np.genfromtxt(FileName, dtype=str,
            delimiter='\n', comments='!')

        # Get mesh dimensions
        dim = np.array(fileLines[0].split('!')[0].split(), dtype=int)
        dim = (dim[0]+1, dim[1]+1, dim[2]+1)

        # The origin corner (Southwest-top)
        #- Remember UBC format specifies down as the positive Z
        #- Easting, Northing, Altitude
        oo = np.array(
            fileLines[1].split('!')[0].split(),
            dtype=float
        )
        ox,oy,oz = oo[0],oo[1],oo[2]

        # Read cell sizes for each line in the UBC mesh files
        def _readCellLine(line):
            line_list = []
            for seg in line.split():
                if '*' in seg:
                    sp = seg.split('*')
                    seg_arr = np.ones((int(sp[0]),), dtype=float) * float(sp[1])
                else:
                    seg_arr = np.array([float(seg)], dtype=float)
                line_list.append(seg_arr)
            return np.concatenate(line_list)

        # Read the cell sizes
        cx = _readCellLine(fileLines[2].split('!')[0])
        cy = _readCellLine(fileLines[3].split('!')[0])
        cz = _readCellLine(fileLines[4].split('!')[0])
        # Invert the indexing of the vector to start from the bottom.
        cz = cz[::-1]
        # Adjust the reference point to the bottom south west corner
        oz = oz - np.sum(cz)

        # Now generate the coordinates for from cell width and origin
        cox = ox + np.cumsum(cx)
        cox = np.insert(cox,0,ox)
        coy = oy + np.cumsum(cy)
        coy = np.insert(coy,0,oy)
        coz = oz + np.cumsum(cz)
        coz = np.insert(coz,0,oz)

        # Set the dims and coordinates for the output
        output.SetDimensions(dim[0],dim[1],dim[2])
        # Convert to VTK array for setting coordinates
        output.SetXCoordinates(nps.numpy_to_vtk(num_array=cox,deep=True))
        output.SetYCoordinates(nps.numpy_to_vtk(num_array=coy,deep=True))
        output.SetZCoordinates(nps.numpy_to_vtk(num_array=coz,deep=True))

        return output


    def __ubcMeshData3D(self, FileName_Mesh, FileName_Models, output):
        """Helper method to read a 3D mesh"""
        # Construct/read the mesh
        if self.NeedToReadMesh():
            TensorMeshReader.ubcMesh3D(FileName_Mesh, self.__mesh)
            self.NeedToReadMesh(flag=False)
        output.DeepCopy(self.__mesh)
        if self.NeedToReadModels() and self.ThisHasModels():
            self.__models = []
            for f in FileName_Models:
                # Read the model data
                self.__models.append(TensorMeshReader.ubcModel3D(f))
            self.NeedToReadModels(flag=False)
        return output


    def __ubcTensorMesh(self, FileName_Mesh, FileName_Models, output):
        """Wrapper to Read UBC GIF 2D and 3D meshes. UBC Mesh 2D/3D models are defined using a 2-file format. The "mesh" file describes how the data is descritized. The "model" file lists the physical property values for all cells in a mesh. A model file is meaningless without an associated mesh file. If the mesh file is 2D, then then model file must also be in the 2D format (same for 3D).

        Args:
            FileName_Mesh (str) : The mesh filename as an absolute path for the input mesh file in UBC 2D/3D Mesh Format
            FileName_Models (str or list(str)) : The model filename(s) as an absolute path for the input model file in UBC 2D/3D Model Format.
            output (vtkRectilinearGrid) : The output data object

        Return:
            vtkRectilinearGrid : Returns a ``vtkRectilinearGrid`` generated from the UBC 2D/3D Mesh grid. Mesh is defined by the input mesh file. Cell data is defined by the input model file.

        """
        # Check if the mesh is a UBC 2D mesh
        if self.Is2D():
            self.__ubcMeshData2D(FileName_Mesh, FileName_Models, output)
        # Check if the mesh is a UBC 3D mesh
        elif self.Is3D():
            self.__ubcMeshData3D(FileName_Mesh, FileName_Models, output)
        else:
            raise _helpers.PVGeoError('File format not recognized')
        return output

    def RequestData(self, request, inInfo, outInfo):
        """Handles data request by the pipeline.
        """
        # Get output:
        output = self.GetOutputData(outInfo, 0)
        # Get requested time index
        i = _helpers.GetRequestedTime(self, outInfo)
        self.__ubcTensorMesh(
            self.GetMeshFileName(),
            self.GetModelFileNames(),
            output)
        # Place the model data for given timestep onto the mesh
        if len(self.__models) > i:
            TensorMeshReader.PlaceModelOnMesh(output, self.__models[i], self.GetDataName())
        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        """Handles info request by pipeline about timesteps and grid extents.
        """
        # Call parent to handle time stuff
        ubcMeshReaderBase.RequestInformation(self, request, inInfo, outInfo)
        # Now set whole output extent
        if self.NeedToReadMesh():
            ext = self._ReadExtent()
            info = outInfo.GetInformationObject(0)
            # Set WHOLE_EXTENT: This is absolutely necessary
            info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), ext, 6)
        return 1

    def ClearMesh(self):
        """Use to clean/rebuild the mesh
        """
        self.__mesh = vtk.vtkRectilinearGrid()
        ubcMeshReaderBase.ClearModels(self)

    def ClearModels(self):
        """Use to clean the models and reread
        """
        self.__models = []
        ubcMeshReaderBase.ClearModels(self)

################################################################################

class TensorMeshAppender(ModelAppenderBase):
    """This filter reads a timeseries of models and appends it to an input
    ``vtkRectilinearGrid``
    """
    __displayname__ = 'UBC Tensor Mesh Appender'
    __type__ = 'filter'
    def __init__(self, **kwargs):
        ModelAppenderBase.__init__(self,
            inputType='vtkRectilinearGrid',
            outputType='vtkRectilinearGrid',
            **kwargs)


    def _ReadUpFront(self):
        reader = ubcMeshReaderBase.ubcModel3D
        if not self._is3D:
            # Note how in UBC format, 2D grids are specified on an XZ plane (no Y component)
            # This will only work prior to rotations to account for real spatial reference
            reader = TensorMeshReader.ubcModel2D
        self._models = []
        for f in self._modelFileNames:
            # Read the model data
            self._models.append(reader(f))
        self.NeedToRead(flag=False)
        return

    def _PlaceOnMesh(self, output, idx=0):
        TensorMeshReader.PlaceModelOnMesh(output, self._models[idx], self._dataname)
        return



paraview_plugin_version = '1.1.10'
# This is module to import. It provides VTKPythonAlgorithmBase, the base class
# for all python-based vtkAlgorithm subclasses in VTK and decorators used to
# 'register' the algorithm with ParaView along with information about UI.
from paraview.util.vtkAlgorithm import *

# Helpers:
from PVGeo import _helpers
# Classes to Decorate
from PVGeo.grids import ReverseImageDataAxii, TranslateGridOrigin, TableToGrid, ExtractTopography


#### GLOBAL VARIABLES ####
MENU_CAT = 'PVGeo: General Grids'


###############################################################################


@smproxy.filter(name='PVGeoReverseImageDataAxii', label='Reverse Image Data Axii')
@smhint.xml('<ShowInMenu category="%s"/>' % MENU_CAT)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkImageData"], composite_data_supported=False)
class PVGeoReverseImageDataAxii(ReverseImageDataAxii):
    def __init__(self):
        ReverseImageDataAxii.__init__(self)

    #### Seters and Geters ####

    @smproperty.xml(_helpers.getPropertyXml(name='Flip X Axis', command='SetFlipX', default_values=True, help='A boolean to set whether to flip the X axis.'))
    def SetFlipX(self, flag):
        ReverseImageDataAxii.SetFlipX(self, flag)

    @smproperty.xml(_helpers.getPropertyXml(name='Flip Y Axis', command='SetFlipY', default_values=True, help='A boolean to set whether to flip the Y axis.'))
    def SetFlipY(self, flag):
        ReverseImageDataAxii.SetFlipY(self, flag)

    @smproperty.xml(_helpers.getPropertyXml(name='Flip Z Axis', command='SetFlipZ', default_values=True, help='A boolean to set whether to flip the Z axis.'))
    def SetFlipZ(self, flag):
        ReverseImageDataAxii.SetFlipZ(self, flag)


###############################################################################


@smproxy.filter(name='PVGeoTranslateGridOrigin', label='Translate Grid Origin')
@smhint.xml('<ShowInMenu category="%s"/>' % MENU_CAT)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkImageData"], composite_data_supported=False)
class PVGeoTranslateGridOrigin(TranslateGridOrigin):
    def __init__(self):
        TranslateGridOrigin.__init__(self)

    #### Seters and Geters ####

    @smproperty.xml(_helpers.getDropDownXml(name='Corner', command='SetCorner',
        labels=['South East Bottom', 'North West Bottom', 'North East Bottom',
        'South West Top', 'South East Top', 'North West Top', 'North East Top'],
        values=[1,2,3,4,5,6,7]))
    def SetCorner(self, corner):
        TranslateGridOrigin.SetCorner(self, corner)


###############################################################################


@smproxy.filter(name='PVGeoTableToGrid', label='Table To Grid')
@smhint.xml('<ShowInMenu category="%s"/>' % MENU_CAT)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkTable"], composite_data_supported=False)
@smhint.xml('''<RepresentationType view="RenderView" type="Surface With Edges" />''')
class PVGeoTableToGrid(TableToGrid):
    def __init__(self):
        TableToGrid.__init__(self)


    #### Setters / Getters ####


    @smproperty.intvector(name="Extent", default_values=[10, 10, 10])
    def SetExtent(self, nx, ny, nz):
        TableToGrid.SetExtent(self, nx, ny, nz)

    @smproperty.doublevector(name="Spacing", default_values=[1.0, 1.0, 1.0])
    def SetSpacing(self, dx, dy, dz):
        TableToGrid.SetSpacing(self, dx, dy, dz)

    @smproperty.doublevector(name="Origin", default_values=[0.0, 0.0, 0.0])
    def SetOrigin(self, x0, y0, z0):
        TableToGrid.SetOrigin(self, x0, y0, z0)

    @smproperty.xml(_helpers.getPropertyXml(name='SEPlib', command='SetSEPlib', default_values=False, help='Use the Stanford Exploration Project\'s axial conventions (d1=z, d2=x, d3=y). Parameters would be entered [z,x,y].'))
    def SetSEPlib(self, flag):
        TableToGrid.SetSEPlib(self, flag)

    @smproperty.xml(_helpers.getDropDownXml(name='Order', command='SetOrder',
        labels=['Fortran-style: column-major order', 'C-style: Row-major order'],
        values=[0, 1]))
    def SetOrder(self, order):
        o = ['F', 'C']
        TableToGrid.SetOrder(self, o[order])

    @smproperty.xml(_helpers.getPropertyXml(name='Swap XY', command='SetSwapXY', default_values=False))
    def SetSwapXY(self, flag):
        TableToGrid.SetSwapXY(self, flag)



###############################################################################


@smproxy.filter(name='PVGeoExtractTopography', label='Extract Topography')
@smhint.xml('<ShowInMenu category="%s"/>' % MENU_CAT)
@smproperty.input(name="Topography", port_index=1)
@smdomain.datatype(dataTypes=["vtkPolyData"], composite_data_supported=False)
@smproperty.input(name="Data Set", port_index=0)
@smdomain.datatype(dataTypes=["vtkDataSet"], composite_data_supported=False)
class PVGeoExtractTopography(ExtractTopography):
    def __init__(self):
        ExtractTopography.__init__(self)

    #### Seters and Geters ####



"""The ``export`` module privides a few macros for exporting data scenes within
ParaView's rendering environment to the VTKjs format.

"""

__all__ = [
    'exportVTKjs',
    'getVTKjsURL',
]

import sys

def exportVTKjs(FileName='', compress=False):
    """This function will execute a script to export the current scene from your rendering into the VTKjs shareable file format.

    Args:
        FileName (str) : Use to specify the basename of the output file. Extension will always be '.vtkjs'
        compress (bool) : Declares a preference to compress the data arrays. Defaults to False.

    Return:
        None: Prints the path to the saved '.vtkjs' file.

    Note:
        - To view, open the file in the VTKjs standalone web viewer `found here`_
        - Use the ``get_vtkjs_url.py`` script in the ``PVGeo`` repository to get a shareable link for the exported file.

    .. _found here: http://viewer.pvgeo.org

    """
    import os
    import sys
    path = os.path.dirname(os.path.abspath(__file__))
    # TODO: debug why compression sometimes fails
    sys.argv = ['%s/%s' % (path, '_export-scene-macro.py'), FileName, False]
    execfile('%s/%s' % (path, '_export-scene-macro.py'), globals())
    return

exportVTKjs.__displayname__ = 'Export VTKjs'
exportVTKjs.__type__ = 'macro'

################################################################################


def convertDropboxURL(url):
    return url.replace("https://www.dropbox.com", "https://dl.dropbox.com")

def convertGitHubURL(url):
    url = url.replace("https://github.com", "https://rawgit.com")
    url = url.replace("raw/", "")
    return url

def generateViewerURL(dataURL):
    viewerURL = "http://viewer.pvgeo.org/"
    return viewerURL + '%s%s' % ("?fileURL=", dataURL)

def getVTKjsURL(host, inURL):
    """After using ``exportVTKjs()`` to create a ``.vtkjs`` file from a ParaView
    data scene which is uploaded to an online file hosting service like Dropbix,
    use this method to get a shareable link to that scene on the
    `PVGeo VTKjs viewer`_.

    .. _PVGeo VTKjs viewer: http://viewer.pvgeo.org

    **Current file hosts supported:**
    - Dropbox
    - GitHub

    Args:
        host (str): the name of the file hosting service.
        inURL (str): the web URL to the ``.vtkjs`` file.

    Example:
        >>> import pvmacros as pvm
        >>> # A Dropbox hosted file:
        >>> pvm.export.getVTKjsURL('dropbox', 'https://www.dropbox.com/s/6m5ttdbv5bf4ngj/ripple.vtkjs\?dl\=0')

        >>> # A GitHib hosted file:
        >>> pvm.export.getVTKjsURL('github', 'https://github.com/OpenGeoVis/PVGeo/raw/docs/ripple.vtkjs')

    """
    if host.lower() == "dropbox":
        convertURL = convertDropboxURL(inURL)
    elif host.lower() == "github":
        convertURL = convertGitHubURL(inURL)
    else:
        print("--> Warning: Web host not specified or supported. URL is simply appended to standalone scene loader link.")
        convertURL = inURL
    #print("--> Your link: %s" % generateViewerURL(convertURL))
    return generateViewerURL(convertURL)

getVTKjsURL.__displayname__ = 'Get VTKjs URL'
getVTKjsURL.__type__ = 'macro'



__all__ = [
    'PVGeoError',
    'ErrorObserver',
]

import re


class PVGeoError(Exception):
    """This is a custom error class for handling errors when proccessing on the VTK pipeline. It makes the error messages easy to decipher in ParaView and cleans the messages when used in Python outside of ParaView. When on the VTK pipeline, errors aren't really raised but passed over and printed to the console. This class makes decipher the error streams a whole lot easier for human eyes.
    """
    QUALIFIER_L = '@@@@PVGeoError ---> '
    QUALIFIER_R = ' <--- PVGeoError@@@@'
    SEARCHER = re.compile(r'@@@@PVGeoError --->.+?<--- PVGeoError@@@@', re.MULTILINE|re.DOTALL)


    def __init__(self, message):
        # Place special characters arround the message for easy extraction
        self.message = '\n\n\n\n' + self.QUALIFIER_L + message + self.QUALIFIER_R + '\n\n\n\n'


    def __str__(self):
        return self.message #.replace(self.QUALIFIER, '')

    @staticmethod
    def CleanMessage(message):
        return message.replace(PVGeoError.QUALIFIER_L, '').replace(PVGeoError.QUALIFIER_R, '')



class ErrorObserver:
    """A class for catching errors when processing on a VTK pipeline. The ``AlgorithmBase`` class handles setting up this observer on initialization.

    Example:
        >>> import PVGeo
        >>> # Only use this observer on sub classes of the AlgorithmBase:
        >>> f = PVGeo.AlgorithmBase()
        >>> f.Update()
        >>> if f.ErrorOccurred():
        >>>    print(f.ErrorMessage())
        ERROR: ...

    """
    def __init__(self):
        self.__ErrorOccurred = False
        self.__ErrorMessage = None
        self.__ErrorMessageEtc = None
        self.CallDataType = 'string0'
        # Object to observe:
        self.__observing = False

    def __call__(self, obj, event, message):
        self.__ErrorOccurred = True
        # Serch the message for a PVGeoError qualifier to extract
        msg = PVGeoError.SEARCHER.findall(message)
        if len(msg) > 0:
            info = '\nPVGeoError: '
            message = info + info.join(PVGeoError.CleanMessage(m) for m in msg)
        elif self.__ErrorMessage is not None:
            self.__ErrorMessageEtc = message
            return
        # if no qualifier is present and message has not already been set, entire message stream gets set
        self.__ErrorMessage = message
        print(message)

    def ErrorOccurred(self):
        """Ask self if an error has occured
        """
        occ = self.__ErrorOccurred
        self.__ErrorOccurred = False
        return occ

    def ErrorMessage(self, etc=False):
        """Get the last set error message

        Return:
            str: the last set error message
        """
        if etc:
            return self.__ErrorMessageEtc
        return self.__ErrorMessage

    def MakeObserver(self, algorithm):
        """Make this an observer of an algorithm
        """
        if self.__observing:
            raise RuntimeError('This error observer is already observing an algorithm.')
        algorithm.GetExecutive().AddObserver('ErrorEvent', self)
        algorithm.AddObserver('ErrorEvent', self)
        self.__observing = True
        return



from . import export
from . import vis
from .pipeline import *

__author__ = 'Bane Sullivan'
__license__ = 'BSD-3-Clause'
__copyright__ = '2018, Bane Sullivan'
__version__ = '1.1.10'


__displayname__ = 'ParaView Macros'
